import os
import sys
from itertools import cycle
from pathlib import Path

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import plotly.express as px

from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from sklearn_genetic.utils import logbook_to_pandas
from sklearn.metrics import ConfusionMatrixDisplay

try:
    from . import misc
except (ModuleNotFoundError, ValueError, ImportError):
    from utils import misc


class Plotting:
    """Functions for plotting imputer scoring and results."""

    @staticmethod
    def plot_grid_search(cv_results, nn_method, prefix):
        """Plot cv_results\_ from a grid search for each parameter.

        Saves a figure to disk.

        Args:
            cv_results (np.ndarray): the cv_results\_ attribute from a trained grid search object.

            nn_method (str): Neural network algorithm name.

            prefix (str): Prefix to use for saving the plot to file.
        """
        ## Results from grid search
        results = pd.DataFrame(cv_results)
        means_test = [col for col in results if col.startswith("mean_test_")]
        filter_col = [col for col in results if col.startswith("param_")]
        params_df = results[filter_col].astype(str)
        for i, col in enumerate(means_test):
            params_df[col] = results[means_test[i]]

        # Get number of needed subplot rows.
        tot = len(filter_col)
        cols = 4
        rows = int(np.ceil(tot / cols))

        fig = plt.figure(1, figsize=(20, 10))
        fig.tight_layout(pad=3.0)

        # Set font properties.
        font = {"size": 12}
        plt.rc("font", **font)

        for i, p in enumerate(filter_col, start=1):
            ax = fig.add_subplot(rows, cols, i)

            # Plot each metric.
            for col in means_test:
                # Get maximum score for each parameter setting.
                df_plot = params_df.groupby(p)[col].agg("max")

                # Convert to float if not supposed to be string.
                try:
                    df_plot.index = df_plot.index.astype(float)
                except TypeError:
                    pass

                # Sort by index (numerically if possible).
                df_plot = df_plot.sort_index()

                # Remove prefix from score name.
                col_new_name = col[len("mean_test_") :]

                ax.plot(
                    df_plot.index.astype(str),
                    df_plot.values,
                    "-o",
                    label=col_new_name,
                )

                ax.legend(loc="best")

            param_new_name = p[len("param_") :]
            ax.set_xlabel(param_new_name.lower())
            ax.set_ylabel("Max Score")
            ax.set_ylim([0, 1])

        fig.savefig(
            os.path.join(
                f"{prefix}_output",
                "plots",
                "Unsupervised",
                nn_method,
                "gridsearch_metrics.pdf",
            ),
            bbox_inches="tight",
            facecolor="white",
        )

    @staticmethod
    def plot_metrics(metrics, num_classes, prefix, nn_method):
        """Plot AUC-ROC and Precision-Recall performance metrics for neural network classifier.

        Saves plot to PDF file on disk.

        Args:
            metrics (Dict[str, Any]): Per-class, micro, and macro-averaged metrics including accuracy, ROC-AUC, and Precision-Recall with Average Precision scores.

            num_classes (int): Number of classes evaluated.

            prefix (str): Prefix to use for output plot.

            nn_method (str): Neural network algorithm being used.
        """
        # Set font properties.
        font = {"size": 12}
        plt.rc("font", **font)

        fn = os.path.join(
            f"{prefix}_output",
            "plots",
            "Unsupervised",
            nn_method,
            f"auc_pr_curves.pdf",
        )
        fig = plt.figure(figsize=(20, 10))

        acc = round(metrics["accuracy"] * 100, 2)
        ham = round(metrics["hamming"], 2)

        fig.suptitle(
            f"Performance Metrics\nAccuracy: {acc}\nHamming Loss: {ham}"
        )
        axs = fig.subplots(nrows=1, ncols=2)
        plt.subplots_adjust(hspace=0.5)

        # Line weight
        lw = 2

        roc_auc = metrics["roc_auc"]
        pr_ap = metrics["precision_recall"]

        metric_list = [roc_auc, pr_ap]

        for metric, ax in zip(metric_list, axs):
            if "fpr_micro" in metric:
                prefix1 = "fpr"
                prefix2 = "tpr"
                lab1 = "ROC"
                lab2 = "AUC"
                xlab = "False Positive Rate"
                ylab = "True Positive Rate"
                title = "Receiver Operating Characteristic (ROC)"
                baseline = [0, 1]

            elif "recall_micro" in metric:
                prefix1 = "recall"
                prefix2 = "precision"
                lab1 = "Precision-Recall"
                lab2 = "AP"
                xlab = "Recall"
                ylab = "Precision"
                title = "Precision-Recall"
                baseline = [metric["baseline"], metric["baseline"]]

                # Plot iso-f1 curves.
                f_scores = np.linspace(0.2, 0.8, num=4)
                for i, f_score in enumerate(f_scores):
                    x = np.linspace(0.01, 1)
                    y = f_score * x / (2 * x - f_score)
                    ax.plot(
                        x[y >= 0],
                        y[y >= 0],
                        color="gray",
                        alpha=0.2,
                        linewidth=lw,
                        label="Iso-F1 Curves" if i == 0 else "",
                    )
                    ax.annotate(f"F1={f_score:0.1f}", xy=(0.9, y[45] + 0.02))

            # Plot ROC curves.
            ax.plot(
                metric[f"{prefix1}_micro"],
                metric[f"{prefix2}_micro"],
                label=f"Micro-averaged {lab1} Curve ({lab2} = {metric['micro']:.2f})",
                color="deeppink",
                linestyle=":",
                linewidth=4,
            )

            ax.plot(
                metric[f"{prefix1}_macro"],
                metric[f"{prefix2}_macro"],
                label=f"Macro-averaged {lab1} Curve ({lab2} = {metric['macro']:.2f})",
                color="navy",
                linestyle=":",
                linewidth=4,
            )

            colors = cycle(["aqua", "darkorange", "cornflowerblue"])
            for i, color in zip(range(num_classes), colors):
                if f"{prefix1}_{i}" in metric:
                    ax.plot(
                        metric[f"{prefix1}_{i}"],
                        metric[f"{prefix2}_{i}"],
                        color=color,
                        lw=lw,
                        label=f"{lab1} Curve of class {i} ({lab2} = {metric[i]:.2f})",
                    )

            if "fpr_micro" in metric:
                # Make center baseline
                ax.plot(
                    baseline,
                    baseline,
                    "k--",
                    linewidth=lw,
                    label="No Classification Skill",
                )
            else:
                ax.plot(
                    [0, 1],
                    baseline,
                    "k--",
                    linewidth=lw,
                    label="No Classification Skill",
                )

            ax.set_xlim(0.0, 1.0)
            ax.set_ylim(0.0, 1.05)
            ax.set_xlabel(f"{xlab}")
            ax.set_ylabel(f"{ylab}")
            ax.set_title(f"{title}")
            ax.legend(loc="best")

        fig.savefig(fn, bbox_inches="tight", facecolor="white")
        plt.close()
        plt.clf()
        plt.cla()

    @staticmethod
    def plot_search_space(
        estimator,
        height=2,
        s=25,
        features=None,
    ):
        """Make density and contour plots for showing search space during grid search.

        Modified from sklearn-genetic-opt function to implement exception handling.

        Args:
            estimator (sklearn estimator object): A fitted estimator from :class:`~sklearn_genetic.GASearchCV`.

            height (float, optional): Height of each facet. Defaults to 2.

            s (float, optional): Size of the markers in scatter plot. Defaults to 5.

            features (list, optional): Subset of features to plot, if ``None`` it plots all the features by default. Defaults to None.

        Returns:
            g (seaborn.PairGrid): Pair plot of the used hyperparameters during the search.
        """
        sns.set_style("white")

        df = logbook_to_pandas(estimator.logbook)
        if features:
            _stats = df[features]
        else:
            variables = [*estimator.space.parameters, "score"]
            _stats = df[variables]

        g = sns.PairGrid(_stats, diag_sharey=False, height=height)

        g = g.map_upper(sns.scatterplot, s=s, color="r", alpha=0.2)

        try:
            g = g.map_lower(
                sns.kdeplot,
                shade=True,
                cmap=sns.color_palette("ch:s=.25,rot=-.25", as_cmap=True),
            )
        except np.linalg.LinAlgError as err:
            if "singular matrix" in str(err).lower():
                g = g.map_lower(sns.scatterplot, s=s, color="b", alpha=1.0)
            else:
                raise

        try:
            g = g.map_diag(
                sns.kdeplot,
                shade=True,
                palette="crest",
                alpha=0.2,
                color="red",
            )
        except np.linalg.LinAlgError as err:
            if "singular matrix" in str(err).lower():
                g = g.map_diag(sns.histplot, color="red", alpha=1.0, kde=False)

        return g

    @staticmethod
    def visualize_missingness(
        genotype_data,
        df,
        zoom=True,
        prefix="imputer",
        horizontal_space=0.6,
        vertical_space=0.6,
        bar_color="gray",
        heatmap_palette="magma",
        plot_format="pdf",
        dpi=300,
    ):
        """Make multiple plots to visualize missing data.

        Args:
            genotype_data (GenotypeData): Initialized GentoypeData object.

            df (pandas.DataFrame): DataFrame with snps to visualize.

            zoom (bool, optional): If True, zooms in to the missing proportion range on some of the plots. If False, the plot range is fixed at [0, 1]. Defaults to True.

            prefix (str, optional): Prefix for output directory and files. Plots and files will be written to a directory called <prefix>_reports. The report directory will be created if it does not already exist. If prefix is None, then the reports directory will not have a prefix. Defaults to 'imputer'.

            horizontal_space (float, optional): Set width spacing between subplots. If your plot are overlapping horizontally, increase horizontal_space. If your plots are too far apart, decrease it. Defaults to 0.6.

            vertical_space (float, optioanl): Set height spacing between subplots. If your plots are overlapping vertically, increase vertical_space. If your plots are too far apart, decrease it. Defaults to 0.6.

            bar_color (str, optional): Color of the bars on the non-stacked barplots. Can be any color supported by matplotlib. See matplotlib.pyplot.colors documentation. Defaults to 'gray'.

            heatmap_palette (str, optional): Palette to use for heatmap plot. Can be any palette supported by seaborn. See seaborn documentation. Defaults to 'magma'.

            plot_format (str, optional): Format to save plots. Can be any of the following: "pdf", "png", "svg", "ps", "eps". Defaults to "pdf".

            dpi (int): The resolution in dots per inch. Defaults to 300.

        Returns:
            pandas.DataFrame: Per-locus missing data proportions.
            pandas.DataFrame: Per-individual missing data proportions.
            pandas.DataFrame: Per-population + per-locus missing data proportions.
            pandas.DataFrame: Per-population missing data proportions.
            pandas.DataFrame: Per-individual and per-population missing data proportions.
        """

        loc, ind, poploc, poptotal, indpop = genotype_data.calc_missing(df)

        ncol = 3
        nrow = 1 if genotype_data.pops is None else 2

        fig, axes = plt.subplots(nrow, ncol, figsize=(8, 11))
        plt.subplots_adjust(wspace=horizontal_space, hspace=vertical_space)
        fig.suptitle("Missingness Report")

        ax = axes[0, 0]

        ax.set_title("Per-Individual")
        ax.barh(genotype_data.samples, ind, color=bar_color, height=1.0)
        if not zoom:
            ax.set_xlim([0, 1])
        ax.set_ylabel("Sample")
        ax.set_xlabel("Missing Prop.")
        ax.tick_params(
            axis="y",
            which="both",
            left=False,
            right=False,
            labelleft=False,
        )

        ax = axes[0, 1]

        ax.set_title("Per-Locus")
        ax.barh(
            range(genotype_data.num_snps), loc, color=bar_color, height=1.0
        )
        if not zoom:
            ax.set_xlim([0, 1])
        ax.set_ylabel("Locus")
        ax.set_xlabel("Missing Prop.")
        ax.tick_params(
            axis="y",
            which="both",
            left=False,
            right=False,
            labelleft=False,
        )

        id_vars = ["SampleID"]
        if poptotal is not None:
            ax = axes[0, 2]

            ax.set_title("Per-Population Total")
            ax.barh(poptotal.index, poptotal, color=bar_color, height=1.0)
            if not zoom:
                ax.set_xlim([0, 1])
            ax.set_xlabel("Missing Prop.")
            ax.set_ylabel("Population")

            ax = axes[1, 0]

            ax.set_title("Per-Population + Per-Locus")
            npops = len(poploc.columns)

            vmax = None if zoom else 1.0

            sns.heatmap(
                poploc,
                vmin=0.0,
                vmax=vmax,
                cmap=sns.color_palette(heatmap_palette, as_cmap=True),
                yticklabels=False,
                cbar_kws={"label": "Missing Prop."},
                ax=ax,
            )
            ax.set_xlabel("Population")
            ax.set_ylabel("Locus")

            id_vars.append("Population")

        melt_df = indpop.isna()
        melt_df["SampleID"] = genotype_data.samples
        indpop["SampleID"] = genotype_data.samples

        if poptotal is not None:
            melt_df["Population"] = genotype_data.pops
            indpop["Population"] = genotype_data.pops

        melt_df = melt_df.melt(value_name="Missing", id_vars=id_vars)
        melt_df.sort_values(by=id_vars[::-1], inplace=True)
        melt_df["Missing"].replace(False, "Present", inplace=True)
        melt_df["Missing"].replace(True, "Missing", inplace=True)

        ax = axes[0, 2] if poptotal is None else axes[1, 1]

        ax.set_title("Per-Individual")
        g = sns.histplot(
            data=melt_df,
            y="variable",
            hue="Missing",
            multiple="fill",
            ax=ax,
        )
        ax.tick_params(
            axis="y",
            which="both",
            left=False,
            right=False,
            labelleft=False,
        )
        g.get_legend().set_title(None)

        if poptotal is not None:
            ax = axes[1, 2]

            ax.set_title("Per-Population")
            g = sns.histplot(
                data=melt_df,
                y="Population",
                hue="Missing",
                multiple="fill",
                ax=ax,
            )
            g.get_legend().set_title(None)

        fig.savefig(
            os.path.join(
                f"{prefix}_output", "plots", f"missingness.{plot_format}"
            ),
            bbox_inches="tight",
            facecolor="white",
        )
        plt.cla()
        plt.clf()
        plt.close()

        return loc, ind, poploc, poptotal, indpop

    @staticmethod
    def run_and_plot_pca(
        original_genotype_data,
        imputer_object,
        prefix="imputer",
        n_components=3,
        center=True,
        scale=False,
        n_axes=2,
        point_size=15,
        font_size=15,
        plot_format="pdf",
        bottom_margin=0,
        top_margin=0,
        left_margin=0,
        right_margin=0,
        width=1088,
        height=700,
    ):
        """Runs PCA and makes scatterplot with colors showing missingness.

        Genotypes are plotted as separate shapes per population and colored according to missingness per individual.

        This function is run at the end of each imputation method, but can be run independently to change plot and PCA parameters such as ``n_axes=3`` or ``scale=True``.

        The imputed and original GenotypeData objects need to be passed to the function as positional arguments.

        PCA (principal component analysis) scatterplot can have either two or three axes, set with the n_axes parameter.

        The plot is saved as both an interactive HTML file and as a static image. Each population is represented by point shapes. The interactive plot has associated metadata when hovering over the points.

        Files are saved to a reports directory as <prefix>_output/imputed_pca.<plot_format|html>. Supported image formats include: "pdf", "svg", "png", and "jpeg" (or "jpg").

        Args:
            original_genotype_data (GenotypeData): Original GenotypeData object that was input into the imputer.

            imputer_object (Any imputer instance): Imputer object created when imputing. Can be any of the imputers, such as: ``ImputePhylo()``, ``ImputeUBP()``, and ``ImputeRandomForest()``.

            original_012 (pandas.DataFrame, numpy.ndarray, or List[List[int]], optional): Original 012-encoded genotypes (before imputing). Missing values are encoded as -9. This object can be obtained as ``df = GenotypeData.genotypes012_df``.

            prefix (str, optional): Prefix for report directory. Plots will be save to a directory called <prefix>_output/imputed_pca<html|plot_format>. Report directory will be created if it does not already exist. Defaults to "imputer".

            n_components (int, optional): Number of principal components to include in the PCA. Defaults to 3.

            center (bool, optional): If True, centers the genotypes to the mean before doing the PCA. If False, no centering is done. Defaults to True.

            scale (bool, optional): If True, scales the genotypes to unit variance before doing the PCA. If False, no scaling is done. Defaults to False.

            n_axes (int, optional): Number of principal component axes to plot. Must be set to either 2 or 3. If set to 3, a 3-dimensional plot will be made. Defaults to 2.

            point_size (int, optional): Point size for scatterplot points. Defaults to 15.

            plot_format (str, optional): Plot file format to use. Supported formats include: "pdf", "svg", "png", and "jpeg" (or "jpg"). An interactive HTML file is also created regardless of this setting. Defaults to "pdf".

            bottom_margin (int, optional): Adjust bottom margin. If whitespace cuts off some of your plot, lower the corresponding margins. The default corresponds to that of plotly update_layout(). Defaults to 0.

            top (int, optional): Adjust top margin. If whitespace cuts off some of your plot, lower the corresponding margins. The default corresponds to that of plotly update_layout(). Defaults to 0.

            left_margin (int, optional): Adjust left margin. If whitespace cuts off some of your plot, lower the corresponding margins. The default corresponds to that of plotly update_layout(). Defaults to 0.

            right_margin (int, optional): Adjust right margin. If whitespace cuts off some of your plot, lower the corresponding margins. The default corresponds to that of plotly update_layout(). Defaults to 0.

            width (int, optional): Width of plot space. If your plot is cut off at the edges, even after adjusting the margins, increase the width and height. Try to keep the aspect ratio similar. Defaults to 1088.

            height (int, optional): Height of plot space. If your plot is cut off at the edges, even after adjusting the margins, increase the width and height. Try to keep the aspect ratio similar. Defaults to 700.

        Returns:
            numpy.ndarray: PCA data as a numpy array with shape (n_samples, n_components).

            sklearn.decomposision.PCA: Scikit-learn PCA object from sklearn.decomposision.PCA. Any of the sklearn.decomposition.PCA attributes can be accessed from this object. See sklearn documentation.

        Examples:
            >>>data = GenotypeData(
            >>>    filename="snps.str",
            >>>    filetype="structure2row",
            >>>    popmapfile="popmap.txt",
            >>>)
            >>>
            >>>ubp = ImputeUBP(genotype_data=data)
            >>>
            >>>components, pca = run_and_plot_pca(
            >>>    data,
            >>>    ubp,
            >>>    scale=True,
            >>>    center=True,
            >>>    plot_format="png"
            >>>)
            >>>
            >>>explvar = pca.explained_variance_ratio\_
        """
        report_path = os.path.join(f"{prefix}_output", "plots")
        Path(report_path).mkdir(parents=True, exist_ok=True)

        if n_axes > 3:
            raise ValueError(
                ">3 axes is not supported; n_axes must be either 2 or 3."
            )
        if n_axes < 2:
            raise ValueError(
                "<2 axes is not supported; n_axes must be either 2 or 3."
            )

        imputer = imputer_object.imputed

        df = misc.validate_input_type(
            imputer.genotypes012_df, return_type="df"
        )

        original_df = misc.validate_input_type(
            original_genotype_data.genotypes_012(fmt="pandas"),
            return_type="df",
        )

        original_df.replace(-9, np.nan, inplace=True)

        if center or scale:
            # Center data to mean. Scaling to unit variance is off.
            scaler = StandardScaler(with_mean=center, with_std=scale)
            pca_df = scaler.fit_transform(df)
        else:
            pca_df = df.copy()

        # Run PCA.
        model = PCA(n_components=n_components)
        components = model.fit_transform(pca_df)

        df_pca = pd.DataFrame(
            components[:, [0, 1, 2]], columns=["Axis1", "Axis2", "Axis3"]
        )

        df_pca["SampleID"] = original_genotype_data.samples
        df_pca["Population"] = original_genotype_data.pops
        df_pca["Size"] = point_size

        _, ind, _, _, _ = imputer.calc_missing(original_df, use_pops=False)
        df_pca["missPerc"] = ind

        my_scale = [("rgb(19, 43, 67)"), ("rgb(86,177,247)")]  # ggplot default

        z = "Axis3" if n_axes == 3 else None
        labs = {
            "Axis1": f"PC1 ({round(model.explained_variance_ratio_[0] * 100, 2)}%)",
            "Axis2": f"PC2 ({round(model.explained_variance_ratio_[1] * 100, 2)}%)",
            "missPerc": "Missing Prop.",
            "Population": "Population",
        }

        if z is not None:
            labs[
                "Axis3"
            ] = f"PC3 ({round(model.explained_variance_ratio_[2] * 100, 2)}%)"
            fig = px.scatter_3d(
                df_pca,
                x="Axis1",
                y="Axis2",
                z="Axis3",
                color="missPerc",
                symbol="Population",
                color_continuous_scale=my_scale,
                custom_data=["Axis3", "SampleID", "Population", "missPerc"],
                size="Size",
                size_max=point_size,
                labels=labs,
            )
        else:
            fig = px.scatter(
                df_pca,
                x="Axis1",
                y="Axis2",
                color="missPerc",
                symbol="Population",
                color_continuous_scale=my_scale,
                custom_data=["Axis3", "SampleID", "Population", "missPerc"],
                size="Size",
                size_max=point_size,
                labels=labs,
            )
        fig.update_traces(
            hovertemplate="<br>".join(
                [
                    "Axis 1: %{x}",
                    "Axis 2: %{y}",
                    "Axis 3: %{customdata[0]}",
                    "Sample ID: %{customdata[1]}",
                    "Population: %{customdata[2]}",
                    "Missing Prop.: %{customdata[3]}",
                ]
            ),
        )
        fig.update_layout(
            showlegend=True,
            margin=dict(
                b=bottom_margin,
                t=top_margin,
                l=left_margin,
                r=right_margin,
            ),
            width=width,
            height=height,
            legend_orientation="h",
            legend_title="Population",
            legend_title_font=dict(size=font_size),
            legend_title_side="top",
            font=dict(size=font_size),
        )
        fig.write_html(os.path.join(report_path, "imputed_pca.html"))
        fig.write_image(
            os.path.join(report_path, f"imputed_pca.{plot_format}"),
        )

        return components, model

    @staticmethod
    def plot_history(lod, nn_method, prefix="imputer"):
        """Plot model history traces. Will be saved to file.

        Args:
            lod (List[tf.keras.callbacks.History]): List of history objects.
            nn_method (str): Neural network method to plot. Possible options include: 'NLPCA', 'UBP', or 'VAE'. NLPCA and VAE get plotted the same, but UBP does it differently due to its three phases.
            prefix (str, optional): Prefix to use for output directory. Defaults to 'imputer'.

        Raises:
            ValueError: nn_method must be either 'NLPCA', 'UBP', or 'VAE'.
        """
        if nn_method == "NLPCA" or nn_method == "VAE" or nn_method == "SAE":
            title = nn_method
            fn = os.path.join(
                f"{prefix}_output",
                "plots",
                "Unsupervised",
                nn_method,
                "histplot.pdf",
            )

            # if nn_method == "VAE":
            fig, axes = plt.subplots(1, 2)
            ax1 = axes[0]
            ax2 = axes[1]

            fig.suptitle(title)
            fig.tight_layout(h_pad=3.0, w_pad=3.0)
            history = lod[0]

            acctrain = "binary_accuracy"
            accval = "val_binary_accuracy"
            lossval = "val_loss"

            # Plot train accuracy
            ax1.plot(history[acctrain])
            ax1.set_title("Model Accuracy")
            ax1.set_ylabel("Accuracy")
            ax1.set_xlabel("Epoch")
            ax1.set_ylim(bottom=0.0, top=1.0)
            ax1.set_yticks([0.0, 0.25, 0.5, 0.75, 1.0])

            labels = ["Train"]

            # Plot validation accuracy
            ax1.plot(history[accval])
            labels.append("Validation")

            ax1.legend(labels, loc="best")

            # else:
            ax2.plot(history["loss"])
            ax2.plot(history[lossval])

            ax2.set_title("Total Loss")
            ax2.set_ylabel("Loss")
            ax2.set_xlabel("Epoch")
            ax2.legend(labels, loc="best")

            fig.savefig(fn, bbox_inches="tight", facecolor="white")

            plt.close()
            plt.clf()

        elif nn_method == "UBP":
            fig = plt.figure(figsize=(12, 16))
            fig.suptitle(nn_method)
            fig.tight_layout(h_pad=2.0, w_pad=2.0)
            fn = os.path.join(
                f"{prefix}_output",
                "plots",
                "Unsupervised",
                nn_method,
                "histplot.pdf",
            )

            idx = 1
            for i, history in enumerate(lod, start=1):
                plt.subplot(3, 2, idx)
                title = f"Phase {i}"

                # Plot model accuracy
                ax = plt.gca()
                ax.plot(history["binary_accuracy"])
                ax.set_title(f"{title} Accuracy")
                ax.set_ylabel("Accuracy")
                ax.set_xlabel("Epoch")
                ax.set_yticks([0.0, 0.25, 0.5, 0.75, 1.0])

                # Plot validation accuracy
                ax.plot(history["val_binary_accuracy"])
                ax.legend(["Train", "Validation"], loc="best")

                # Plot model loss
                plt.subplot(3, 2, idx + 1)
                ax = plt.gca()
                ax.plot(history["loss"])
                ax.set_title(f"{title} Loss")
                ax.set_ylabel("Loss (MSE)")
                ax.set_xlabel("Epoch")

                # Plot validation loss
                ax.plot(history["val_loss"])
                ax.legend(["Train", "Validation"], loc="best")

                idx += 2

            plt.savefig(fn, bbox_inches="tight", facecolor="white")

            plt.close()
            plt.clf()

        else:
            raise ValueError(
                f"nn_method must be either 'NLPCA', 'UBP', or 'VAE', but got {nn_method}"
            )

    @staticmethod
    def plot_certainty_heatmap(
        y_certainty, sample_ids=None, nn_method="VAE", prefix="imputer"
    ):
        fig = plt.figure()
        hm = sns.heatmap(
            data=y_certainty,
            cmap="viridis",
            vmin=0.0,
            vmax=1.0,
            cbar_kws={"label": "Prob."},
        )
        hm.set_xlabel("Site")
        hm.set_ylabel("Sample")
        hm.set_title("Probabilities of Uncertain Sites")
        fig.tight_layout()
        fig.savefig(
            os.path.join(
                f"{prefix}_output",
                "plots",
                "Unsupervised",
                nn_method,
                "uncertainty_plot.png",
            ),
            bbox_inches="tight",
            facecolor="white",
        )

    @staticmethod
    def plot_confusion_matrix(
        y_true_1d, y_pred_1d, nn_method, prefix="imputer"
    ):
        fig, ax = plt.subplots(1, 1, figsize=(15, 15))
        ConfusionMatrixDisplay.from_predictions(
            y_true=y_true_1d, y_pred=y_pred_1d, ax=ax
        )

        outfile = os.path.join(
            f"{prefix}_output",
            "plots",
            "Unsupervised",
            nn_method,
            f"confusion_matrix_{nn_method}.png",
        )

        if os.path.isfile(outfile):
            os.remove(outfile)

        fig.savefig(outfile, facecolor="white")

    @staticmethod
    def plot_gt_distribution(df, plot_path):
        df = misc.validate_input_type(df, return_type="df")
        df_melt = pd.melt(df, value_name="Count")
        cnts = df_melt["Count"].value_counts()
        cnts.index.names = ["Genotype"]
        cnts = pd.DataFrame(cnts).reset_index()
        cnts.sort_values(by="Genotype", inplace=True)
        cnts["Genotype"] = cnts["Genotype"].astype(str)

        fig, ax = plt.subplots(1, 1, figsize=(15, 15))
        g = sns.barplot(x="Genotype", y="Count", data=cnts, ax=ax)
        g.set_xlabel("Integer-encoded Genotype")
        g.set_ylabel("Count")
        g.set_title("Genotype Counts")
        for p in g.patches:
            g.annotate(
                f"{p.get_height():.1f}",
                (p.get_x() + 0.25, p.get_height() + 0.01),
                xytext=(0, 1),
                textcoords="offset points",
                va="bottom",
            )

        fig.savefig(
            os.path.join(plot_path, "genotype_distributions.png"),
            bbox_inches="tight",
            facecolor="white",
        )
        plt.close()

    @staticmethod
    def plot_label_clusters(z_mean, labels, prefix="imputer"):
        """Display a 2D plot of the classes in the latent space."""
        fig, ax = plt.subplots(1, 1, figsize=(15, 15))

        sns.scatterplot(x=z_mean[:, 0], y=z_mean[:, 1], ax=ax)
        ax.set_xlabel("Latent Dimension 1")
        ax.set_ylabel("Latent Dimension 2")

        outfile = os.path.join(
            f"{prefix}_output",
            "plots",
            "Unsupervised",
            "VAE",
            "label_clusters.png",
        )

        if os.path.isfile(outfile):
            os.remove(outfile)

        fig.savefig(outfile, facecolor="white", bbox_inches="tight")



# Standard library imports
import os
from typing import Optional, Union, Dict, Any

# Third-party imports
import numpy as np
import pandas as pd

# Scikit-learn imports
import xgboost as xgb
from sklearn.ensemble import ExtraTreesClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier

# Custom imports
try:
    from .impute import Impute
    from .unsupervised.neural_network_imputers import VAE, UBP, SAE
except (ModuleNotFoundError, ValueError, ImportError):
    from pgsui.impute.impute import Impute
    from pgsui.impute.unsupervised.neural_network_imputers import VAE, UBP, SAE


class UnsupervisedImputer(Impute):
    """Parent class for unsupervised imputers. Contains all common arguments and code between unsupervised imputers.

    Args:

        genotype_data (GenotypeData object): GenotypeData instance that was used to read in the sequence data.

        prefix (str): Prefix for output directory. Defaults to "imputer".

        gridparams (Dict[str, Any] or None, optional): Dictionary with keys=keyword arguments for the specified estimator and values=lists of parameter values or distributions. If ``gridparams=None``\, a grid search is not performed, otherwise ``gridparams`` will be used to specify parameter ranges or distributions for the grid search. If using ``gridsearch_method="gridsearch"``\, then the ``gridparams`` values can be lists or numpy arrays. If using ``gridsearch_method="randomized_gridsearch"``\, distributions can be specified by using scipy.stats.uniform(low, high) (for a uniform distribution) or scipy.stats.loguniform(low, high) (useful if range of values spans orders of magnitude). If using the genetic algorithm grid search by setting ``gridsearch_method="genetic_algorithm"``\, the parameters can be specified as ``sklearn_genetic.space`` objects. The grid search will determine the optimal parameters as those that maximize the scoring metrics. If it takes a long time, run it with a small subset of the data just to find the optimal parameters for the classifier, then run a full imputation using the optimal parameters. Defaults to None (no gridsearch).

        cv (int, optional): Number of cross-validation folds to use with grid search. Defaults to 5.
        
        validation_split (float, optional): Proportion of training dataset to set aside for loss validation during model training. Defaults to 0.2.

        column_subset (int or float, optional): If float is provided, gets the proportion of the dataset to randomly subset for the grid search or validation. Subsets ``int(n_features * column_subset)`` columns and Should be in the range [0, 1]. It can be small if the grid search or validation takes a long time. If int is provided, subset ``column_subset`` columns. Defaults to 1.0.

        epochs (int, optional): Number of epochs (cycles through the data) to run during training. Defaults to 100.

        batch_size (int, optional): Batch size to train the model with. Model training per epoch is performed over multiple subsets of samples (rows) of size ``batch_size``\. Defaults to 32.

        n_components (int, optional): Number of components (latent dimensions) to compress the input features to. Defaults to 3.

        early_stop_gen (int, optional): Only used with the genetic algorithm grid search option. Stop training early if the model sees ``early_stop_gen`` consecutive generations without improvement to the scoring metric. This can save training time by reducing the number of epochs and generations that are performed. Defaults to 25.

        num_hidden_layers (int, optional): Number of hidden layers to use in the model. Adjust if overfitting or underfitting occurs. Defaults to 1.

        hidden_layer_sizes (str, List[int], List[str], or int, optional): Number of neurons to use in the hidden layers. If string or a list of strings is passed, the strings must be either "midpoint", "sqrt", or "log2". "midpoint" will calculate the midpoint as ``(n_features + n_components) / 2``\. If "sqrt" is supplied, the square root of the number of features will be used to calculate the output units. If "log2" is supplied, the units will be calculated as ``log2(n_features)``\. hidden_layer_sizes will calculate and set the number of output units for each hidden layer. If multiple hidden layers are supplied, each subsequent layer's dimensions are further reduced by the "midpoint", "sqrt", or "log2". E.g., if using ``num_hidden_layers=3`` and ``n_components=2``\, and there are 100 features (columns), the hidden layer sizes for ``midpoint`` will be: [51, 27, 14]. If a single string or integer is supplied, the model will use the same number of output units for each hidden layer. If a list of integers or strings is supplied, the model will use the values supplied in the list. The list length must be equal to the ``num_hidden_layers`` and all hidden layer sizes must be > n_components. Defaults to "midpoint".

        hidden_activation (str, optional): The activation function to use for the hidden layers. See tf.keras.activations for more info. Supported activation functions include: ["elu", "selu", "leaky_relu", "prelu", "relu"]. Each activation function has some advantages and disadvantages and determines the curve and non-linearity of gradient descent. Some are also faster than others. See https://towardsdatascience.com/7-popular-activation-functions-you-should-know-in-deep-learning-and-how-to-use-them-with-keras-and-27b4d838dfe6 for more information. Note that using ``hidden_activation="selu"`` will force ``weights_initializer`` to be "lecun_normal". Defaults to "elu".

        optimizer (str, optional): The optimizer to use with gradient descent. Supported options are: "adam", "sgd", and "adagrad". See tf.keras.optimizers for more info. Defaults to "adam".

        learning_rate (float, optional): The learning rate for the optimizer. Adjust if the loss is learning too slowly or quickly. If you are getting overfitting, it is likely too high, and likewise underfitting can occur when the learning rate is too low. Defaults to 0.01.

        lr_patience (int, optional): Number of epochs without loss improvement to wait before reducing the learning rate. Defaults to 1.0.

        weights_initializer (str, optional): Initializer to use for the model weights. See tf.keras.initializers for more info. Defaults to "glorot_normal".

        l1_penalty (float, optional): L1 regularization penalty to apply. Adjust if the model is over or underfitting. If this value is too high, underfitting can occur, and vice versa. Defaults to 1e-6.

        l2_penalty (float, optional) L2 regularization penalty to apply. If this value is too high, underfitting can occur, and vice versa. Defaults to 1e-6.

        dropout_rate (float, optional): Neuron dropout rate during training. Dropout randomly disables ``dropout_rate`` proportion of neurons during training, which can reduce overfitting. E.g., if dropout_rate is set to 0.2, then 20% of the neurons are randomly dropped out per epoch. Adjust if the model is over or underfitting. Must be a float in the range [0, 1]. Defaults to 0.2.

        sample_weights (str, Dict[int, float], or None, optional): Weights for the ACTG-encoded classes during training. If None, then does not weight classes. If set to "auto", then class weights are automatically calculated for each column to balance classes. If a dictionary is passed, it must contain "A", "C", "G", and "T" as the keys and the class weights as the values. E.g., {"A": 1.0, "C": 1.0, "G": 1.0, "T": 1.0}. The dictionary is then used as the overall class weights. Defaults to None (no weighting).

        gridsearch_method (str, optional): Grid search method to use. Supported options include: {"gridsearch", "randomized_gridsearch", "genetic_algorithm"}. "gridsearch" uses GridSearchCV to test every possible parameter combination. "randomized_gridsearch" picks ``grid_iter`` random combinations of parameters to test. "genetic_algorithm" uses a genetic algorithm via the sklearn-genetic-opt GASearchCV module to do the grid search. If doing a grid search, "randomized_search" takes the least amount of time because it does not have to test all parameters. "genetic_algorithm" takes the longest. See the scikit-learn GridSearchCV and RandomizedSearchCV documentation for the "gridsearch" and "randomized_gridsearch" options, and the sklearn-genetic-opt GASearchCV documentation for the "genetic_algorithm" option. Defaults to "gridsearch".

        grid_iter (int, optional): Number of iterations to use for randomized and genetic algorithm grid searches. For randomized grid search, ``grid_iter`` parameter combinations will be randomly sampled. For the genetic algorithm, this determines how many generations the genetic algorithm will run. Defaults to 80.

        scoring_metric (str, optional): Scoring metric to use for grid searches. The neural network imputers use a multimetric scorer and use different string values for the grid searches. Supported options include: {"accuracy", "hamming", "roc_auc_micro", "roc_auc_macro", "roc_auc_weighted", "average_precision_micro", "average_precision_macro", "average_precision_weighted", "f1_micro", "f1_macro", and "f1_weighted"}. All of the above metrics are calculated during the grid search, but the provided string just sets the metric that the grid search refits to (i.e., which one is used in the best estimator). See the scikit-learn documentation (https://scikit-learn.org/stable/modules/model_evaluation.html) for more information. Defaults to "f1_weighted".

        population_size (int or str, optional): Only used for the genetic algorithm grid search. Size of the initial population to sample randomly generated individuals. If set to "auto", then ``population_size`` is calculated as ``15 * n_parameters``\. If set to an integer, then uses the integer value as ``population_size``\. If you need to speed up the genetic algorithm grid search, try decreasing this parameter. See GASearchCV in the sklearn-genetic-opt documentation (https://sklearn-genetic-opt.readthedocs.io) for more info. Defaults to "auto".

        tournament_size (int, optional): For genetic algorithm grid search only. Number of individuals to perform tournament selection. See GASearchCV in the sklearn-genetic-opt documentation (https://sklearn-genetic-opt.readthedocs.io) for more info. Defaults to 3.

        elitism (bool, optional): For genetic algorithm grid search only. If set to True, takes the ``tournament_size`` best solution to the next generation. See GASearchCV in the sklearn-genetic-opt documentation (https://sklearn-genetic-opt.readthedocs.io) for more info. Defaults to True.

        crossover_probability (float, optional): For genetic algorithm grid search only. Probability of crossover operation between two individuals. See GASearchCV in the sklearn-genetic-opt documentation (https://sklearn-genetic-opt.readthedocs.io) for more info. Defaults to 0.2.

        mutation_probability (float, optional): For genetic algorithm grid search only. Probability of child mutation. See GASearchCV in the sklearn-genetic-opt documentation (https://sklearn-genetic-opt.readthedocs.io) for more info. Defaults to 0.8.

        ga_algorithm (str, optional): For genetic algorithm grid search only. Evolutionary algorithm to use. Supported options include: {"eaMuPlusLambda", "eaMuCommaLambda", "eaSimple"}. If you need to speed up the genetic algorithm grid search, try setting ``algorithm`` to "euSimple", at the expense of evolutionary model robustness. See more details in the DEAP algorithms documentation (https://deap.readthedocs.io). Defaults to "eaMuPlusLambda".

        sim_strategy (str, optional): Strategy to use for simulating missing data. Only used to validate the accuracy of the imputation. The final model will be trained with the non-simulated dataset. Supported options include: {"random", "random_weighted", "nonrandom", "nonrandom_weighted"}. "random" randomly simulates missing data, while "random_weighted" also does this but balances selection of reference, heterozygous, and alternate alleles. When set to "nonrandom", branches from ``GenotypeData.guidetree`` will be randomly sampled to generate missing data on descendant nodes. For "nonrandom_weighted", missing data will be placed on nodes proportionally to their branch lengths (e.g., to generate data distributed as might be the case with mutation-disruption of RAD sites). If using the "nonrandom" or "nonrandom_weighted" options, a guide tree is required to have been initialized in the passed ``genotype_data`` object. Defaults to "random".

        sim_prop_missing (float, optional): Proportion of missing data to use with missing data simulation. Defaults to 0.2.

        disable_progressbar (bool, optional): Whether to disable the tqdm progress bar. Useful if you are doing the imputation on e.g. a high-performance computing cluster, where sometimes tqdm does not work correctly when being written to a file. If False, uses tqdm progress bar. If True, does not use tqdm. Defaults to False.

        n_jobs (int, optional): Number of parallel jobs to use in the grid search if ``gridparams`` is not None. -1 means use all available processors. Defaults to -1 (all CPUs).

        verbose (int, optional): Verbosity flag. The higher, the more verbose. Possible values are 0, 1, or 2. 0 = silent, 1 = progress bar, 2 = one line per epoch. Note that the progress bar is not particularly useful when logged to a file, so verbose=0 or verbose=2 is recommended when not running interactively. Setting verbose higher than 0 is useful for initial runs and debugging, but can slow down training. Defaults to 0.

        kwargs (Dict[str, Any], optional): Possible options include: {"testing": True/False}. If testing is True, a confusion matrix plot will be created showing model performance. Arrays of the true and predicted values will also be printed to STDOUT. testing defaults to False.

    Attributes:
        imputed (GenotypeData): New GenotypeData instance with imputed data.

        best_params (Dict[str, Any]): Best found parameters from grid search.
    """

    def __init__(
        self,
        genotype_data,
        clf,
        clf_type,
        *,
        prefix="imputer",
        gridparams=None,
        cv: int=5,
        validation_split=0.2,
        column_subset=1.0,
        epochs=100,
        batch_size=32,
        n_components=3,
        early_stop_gen=25,
        num_hidden_layers=1,
        hidden_layer_sizes="midpoint",
        optimizer="adam",
        hidden_activation="elu",
        learning_rate=0.01,
        weights_initializer="glorot_normal",
        l1_penalty=1e-6,
        l2_penalty=1e-6,
        dropout_rate=0.2,
        kl_beta=1.0,
        sample_weights=None,
        gridsearch_method="gridsearch",
        grid_iter=80,
        scoring_metric="f1_weighted",
        population_size="auto",
        tournament_size=3,
        elitism=True,
        crossover_probability=0.2,
        mutation_probability=0.8,
        ga_algorithm="eaMuPlusLambda",
        sim_strategy="random_weighted",
        sim_prop_missing=0.2,
        disable_progressbar=False,
        n_jobs=-1,
        verbose=0,
        **kwargs,
    ):
        all_kwargs = locals()
        all_kwargs.pop("clf")
        all_kwargs.pop("clf_type")

        self.clf = clf
        self.clf_type = clf_type

        imp_kwargs = {
            "str_encodings": {"A": 1, "C": 2, "G": 3, "T": 4, "N": -9},
        }
        all_kwargs.update(imp_kwargs)
        all_kwargs.pop("kwargs")

        super().__init__(self.clf, self.clf_type, all_kwargs)

        if genotype_data is None:
            raise TypeError("genotype_data cannot be NoneType")

        X = genotype_data.genotypes_int

        if not isinstance(X, pd.DataFrame):
            df = pd.DataFrame(X)
        else:
            df = X.copy()

        self.imputed, self.best_params = self.fit_predict(df)


class SupervisedImputer(Impute):
    """Parent class for the supervised imputers. Contains all common arguments and code between supervised imputers.


    Args:

        genotype_data (GenotypeData object): GenotypeData instance that was used to read in the sequence data.

        prefix (str): Prefix for imputed data's output directory.

        gridparams (Dict[str, Any] or None, optional): Dictionary with keys=keyword arguments for the specified estimator and values=lists of parameter values or distributions. If ``gridparams=None``\, a grid search is not performed, otherwise ``gridparams`` will be used to specify parameter ranges or distributions for the grid search. If using ``gridsearch_method="gridsearch"``\, then the ``gridparams`` values can be lists or numpy arrays. If using ``gridsearch_method="randomized_gridsearch"``\, distributions can be specified by using scipy.stats.uniform(low, high) (for a uniform distribution) or scipy.stats.loguniform(low, high) (useful if range of values spans orders of magnitude). If using the genetic algorithm grid search by setting ``gridsearch_method="genetic_algorithm"``\, the parameters can be specified as ``sklearn_genetic.space`` objects. The grid search will determine the optimal parameters as those that maximize the scoring_methods. NOTE: Takes a long time, so you can run it with a small subset of the data using the ``column_subset`` argument just to find the optimal parameters for the classifier, then it will automatically run a full imputation using the optimal parameters. Defaults to None (no gridsearch).

        do_validation (bool, optional): Whether to validate the imputation if not doing a grid search. This validation method randomly replaces between 15% and 50% of the known, non-missing genotypes in ``n_features * column_subset`` of the features. It then imputes the newly missing genotypes for which we know the true values and calculates validation scores. This procedure is replicated ``cv`` times and a mean, median, minimum, maximum, lower 95% confidence interval (CI) of the mean, and the upper 95% CI are calculated and saved to a CSV file. ``gridparams`` must be set to None for ``do_validation`` to work. Calculating a validation score can be turned off altogether by setting ``do_validation`` to False. Defaults to False.

        column_subset (int or float, optional): If float, proportion of the dataset to randomly subset for the grid search or validation. Should be between 0 and 1, and should also be small, because the grid search or validation takes a long time. If int, subset ``column_subset`` columns. If float, subset ``int(n_features * column_subset)`` columns. Defaults to 0.1.

        cv (int, optional): Number of folds for cross-validation during grid search. Defaults to 5.

        max_iter (int, optional): Maximum number of imputation rounds to perform before returning the imputations computed during the final round. A round is a single imputation of each feature with missing values. Defaults to 10.

        tol (float, optional): Tolerance of the stopping condition for the iterations. Defaults to 1e-3.

        n_nearest_features (int, optional): Number of other features to use to estimate the missing values of eacah feature column. If None, then all features will be used, but this can consume an  intractable amount of computing resources. Nearness between features is measured using the absolute correlation coefficient between each feature pair (after initial imputation). To ensure coverage of features throughout the imputation process, the neighbor features are not necessarily nearest, but are drawn with probability proportional to correlation for each imputed target feature. Reducing this can provide significant speed-up when the number of features is large. Defaults to 10.

        initial_strategy (str, optional): Which strategy to use for initializing the missing values in the training data (neighbor columns). IterativeImputer must initially impute the training data (neighbor columns) using a simple, quick imputation in order to predict the missing values for each target column. The ``initial_strategy`` argument specifies which method to use for this initial imputation. Valid options include: “most_frequent”, "populations", "phylogeny", or "mf". "most_frequent" uses the overall mode of each column. "populations" uses the mode per population/ per column via a population map file and the ``ImputeAlleleFreq`` class. "phylogeny" uses an input phylogenetic tree and a rate matrix with the ``ImputePhylo`` class. "mf" performs the imputaton via matrix factorization with the ``ImputeMF`` class. Note that the "mean" and "median" options from the original IterativeImputer are not supported because they are not sensible settings for the type of input data used here. Defaults to "populations".

        str_encodings (dict(str: int), optional): Integer encodings for nucleotides if input file was in STRUCTURE format. Only used if ``initial_strategy="phylogeny"``\. Defaults to {"A": 1, "C": 2, "G": 3, "T": 4, "N": -9}.

        imputation_order (str, optional): The order in which the features will be imputed. Possible values: "ascending" (from features with fewest missing values to most), "descending" (from features with most missing values to fewest), "roman" (left to right), "arabic" (right to left), "random" (a random order for each round). Defaults to "ascending".

        skip_complete (bool, optional): If True, then features with missing values during transform that did not have any missing values during fit will be imputed with the initial imputation method only. Set to True if you have many features with no missing values at both fit and transform time to save compute time. Defaults to False.

        random_state (int or None, optional): The seed of the pseudo random number generator to use for the iterative imputer. Randomizes selection of etimator features if n_nearest_features is not None or the imputation_order is "random". Use an integer for determinism. If None, then uses a different random seed each time. Defaults to None.

        gridsearch_method (str, optional): Grid search method to use. Supported options include: {"gridsearch", "randomized_gridsearch", and "genetic_algorithm"}. "gridsearch" uses GridSearchCV to test every possible parameter combination. "randomized_gridsearch" picks ``grid_iter`` random combinations of parameters to test. "genetic_algorithm" uses a genetic algorithm via sklearn-genetic-opt GASearchCV to do the grid search. If doing a grid search, "randomized_search" takes the least amount of time because it does not have to test all parameters. "genetic_algorithm" takes the longest. See the scikit-learn GridSearchCV and RandomizedSearchCV documentation for the "gridsearch" and "randomized_gridsearch" options, and the sklearn-genetic-opt GASearchCV documentation (https://sklearn-genetic-opt.readthedocs.io) for the "genetic_algorithm" option. Defaults to "gridsearch".

        grid_iter (int, optional): Number of iterations for randomized and genetic algorithm grid searches. Defaults to 80.

        population_size (int or str, optional): For genetic algorithm grid search: Size of the initial population to sample randomly generated individuals. If set to "auto", then ``population_size`` is calculated as ``15 * n_parameters``\. If set to an integer, then uses the integer value as ``population_size``\. If you need to speed up the genetic algorithm grid search, try decreasing this parameter. See GASearchCV in the sklearn-genetic-opt documentation (https://sklearn-genetic-opt.readthedocs.io). Defaults to "auto".

        tournament_size (int, optional): For genetic algorithm grid search: Number of individuals to perform tournament selection. See GASearchCV documentation. Defaults to 3.

        elitism (bool, optional): For genetic algorithm grid search: If True takes the tournament_size best solution to the next generation. See GASearchCV documentation. Defaults to True.

        crossover_probability (float, optional): For genetic algorithm grid search: Probability of crossover operation between two individuals. See GASearchCV documentation. Defaults to 0.2.

        mutation_probability (float, optional): For genetic algorithm grid search: Probability of child mutation. See GASearchCV documentation. Defaults to 0.8.

        ga_algorithm (str, optional): For genetic algorithm grid search: Evolutionary algorithm to use. Supported options include: {"eaMuPlusLambda", "eaMuCommaLambda", "eaSimple"}. If you need to speed up the genetic algorithm grid search, try setting ``algorithm`` to "euSimple", at the expense of evolutionary model robustness. See more details in the DEAP algorithms documentation (https://deap.readthedocs.io). Defaults to "eaMuPlusLambda".

        early_stop_gen (int, optional): If the genetic algorithm sees ``early_stop_gen`` consecutive generations without improvement in the scoring metric, an early stopping callback is implemented. This saves time by reducing the number of generations the genetic algorithm has to perform. Defaults to 5.

        scoring_metric (str, optional): Scoring metric to use for grid searches. See the classification metrics in the scikit-learn documentation (https://scikit-learn.org/stable/modules/model_evaluation.html) for supported options. Defaults to "f1_weighted".

        chunk_size (int or float, optional): Number of loci for which to perform IterativeImputer at one time. Useful for reducing the memory usage if you are running out of RAM. If integer is specified, selects ``chunk_size`` loci at a time. If a float is specified, selects ``math.ceil(total_loci * chunk_size)`` loci at a time]. Defaults to 1.0 (all features).

        disable_progressbar (bool, optional): Whether or not to disable the tqdm progress bar when doing the imputation. If True, progress bar is disabled, which is useful when running the imputation on e.g. an HPC cluster. If the bar is disabled, a status update will be printed to standard output for each iteration and feature instead. If False, the tqdm progress bar will be used. Defaults to False.

        progress_update_percent (int or None, optional): Print status updates for features every ``progress_update_percent``\%. IterativeImputer iterations will always be printed, but ``progress_update_percent`` involves iteration progress through the features of each IterativeImputer iteration. If None, then does not print progress through features. Defaults to None.

        n_jobs (int, optional): Number of parallel jobs to use. If ``gridparams`` is not None, n_jobs is used for the grid search. Otherwise it is used for the classifier. -1 means using all available processors. Defaults to -1 (all CPUs).

        verbose (int, optional): Verbosity flag, controls the debug messages that are issues as functions are evaluated. The higher, the more verbose. Possible values are 0, 1, or 2. Defaults to 0.

    Attributes:
        imputed (GenotypeData): New GenotypeData instance with imputed data.

        best_params (Dict[str, Any]): Best found parameters from grid search.

    """

    def __init__(
        self,
        genotype_data,
        clf,
        clf_type,
        *,
        prefix: str = "output",
        gridparams: Optional[Dict[str, Any]] = None,
        do_validation: bool = False,
        column_subset: Union[int, float] = 0.1,
        cv: int = 5,
        max_iter: int = 10,
        tol: float = 1e-3,
        n_nearest_features: Optional[int] = 10,
        initial_strategy: str = "most_frequent",
        str_encodings: Dict[str, int] = {
            "A": 1,
            "C": 2,
            "G": 3,
            "T": 4,
            "N": -9,
        },
        imputation_order: str = "ascending",
        skip_complete: bool = False,
        random_state: Optional[int] = None,
        gridsearch_method: str = "gridsearch",
        grid_iter: int = 80,
        population_size: Union[int, str] = "auto",
        tournament_size: int = 3,
        elitism: bool = True,
        crossover_probability: float = 0.2,
        mutation_probability: float = 0.8,
        ga_algorithm: str = "eaMuPlusLambda",
        early_stop_gen: int = 5,
        scoring_metric: str = "f1_weighted",
        chunk_size: Union[int, float] = 1.0,
        disable_progressbar: bool = False,
        progress_update_percent: Optional[int] = None,
        n_jobs: int = -1,
        verbose: int = 0,
        **kwargs,
    ):
        # Get local variables into dictionary object
        all_kwargs = locals()
        model_kwargs = all_kwargs.pop("kwargs")
        all_kwargs.update(model_kwargs)
        all_kwargs.pop("self")
        all_kwargs.pop("clf")
        all_kwargs.pop("clf_type")
        all_kwargs.pop("sample_weights")
        all_kwargs.pop("kwargs")

        super().__init__(self.clf, self.clf_type, all_kwargs)

        self.imputed, self.best_params = self.fit_predict(
            genotype_data.genotypes_012(fmt="pandas")
        )


class ImputeKNN(SupervisedImputer):
    """Does K-Nearest Neighbors Iterative Imputation of missing data. Iterative imputation uses the n_nearest_features to inform the imputation at each feature (i.e., SNP site), using the N most correlated features per site. The N most correlated features are drawn with probability proportional to correlation for each imputed target feature to ensure coverage of features throughout the imputation process.

    Args:
        genotype_data (GenotypeData object): GenotypeData instance that was used to read in the sequence data.

        n_neighbors (int, optional): Number of neighbors to use for K-Nearest Neighbors queries. Defaults to 5.

        weights (str, optional): Weight function used in prediction. Possible values: 'Uniform': Uniform weights with all points in each neighborhood weighted equally; 'distance': Weight points by the inverse of their distance, in this case closer neighbors of a query point will have  a greater influence than neighbors that are further away; 'callable': A user-defined function that accepts an array of distances and returns an array of the same shape containing the weights. Defaults to "distance".

        algorithm (str, optional): Algorithm used to compute the nearest neighbors. Possible values: 'ball_tree', 'kd_tree', 'brute', 'auto'. Defaults to "auto".

        leaf_size (int, optional): Leaf size passed to BallTree or KDTree. This can affect the speed of the construction and query, as well as the memory required to store the tree. The optimal value depends on the nature of the problem. Defaults to 30.

        p (int, optional): Power parameter for the Minkowski metric. When p=1, this is equivalent to using manhattan_distance (l1), and if p=2 it is equivalent to using euclidean distance (l2). For arbitrary p, minkowski_distance (l_p) is used. Defaults to 2.

        metric (str, optional): The distance metric to use for the tree. The default metric is minkowski, and with p=2 this is equivalent to the standard Euclidean metric. See the documentation of sklearn.DistanceMetric for a list of available metrics. If metric is 'precomputed', X is assumed to be a distance matrix and must be square during fit. Defaults to "minkowski".

    Example:
        >>> data = GenotypeData(
        >>>     filename="test.str",
        >>>     filetype="auto",
        >>>     guidetree="test.tre",
        >>>     qmatrix_iqtree="test.iqtree"
        >>> )
        >>>
        >>> # Genetic Algorithm grid_params
        >>> grid_params = {
        >>>     "n_neighbors": Integer(3, 10),
        >>>     "leaf_size": Integer(10, 50),
        >>> }
        >>>
        >>> knn = ImputeKNN(
        >>>     genotype_data=data,
        >>>     gridparams=grid_params,
        >>>     cv=5,
        >>>     gridsearch_method="genetic_algorithm",
        >>>     n_nearest_features=10,
        >>>     n_estimators=100,
        >>>     initial_strategy="phylogeny",
        >>> )
        >>>
        >>> knn_gtdata = knn.imputed
    """

    def __init__(
        self,
        genotype_data: Any,
        *,
        n_neighbors: int = 5,
        weights: str = "distance",
        algorithm: str = "auto",
        leaf_size: int = 30,
        p: int = 2,
        metric: str = "minkowski",
        **kwargs,
    ) -> None:
        all_kwargs = locals()
        kwargs.update(all_kwargs)

        self.clf_type = "classifier"
        self.clf = KNeighborsClassifier

        super().__init__(genotype_data, self.clf, self.clf_type, **kwargs)


class ImputeRandomForest(SupervisedImputer):
    """Does Random Forest or Extra Trees Iterative imputation of missing data. Iterative imputation uses the n_nearest_features to inform the imputation at each feature (i.e., SNP site), using the N most correlated features per site. The N most correlated features are drawn with probability proportional to correlation for each imputed target feature to ensure coverage of features throughout the imputation process.

    Args:
        genotype_data (GenotypeData object): GenotypeData instance that was used to read in the sequence data.

        extra_trees (bool, optional): Whether to use ExtraTreesClassifier (If True) instead of RandomForestClassifier (If False). ExtraTreesClassifier is faster, but is not supported by the scikit-learn-intelex patch, whereas RandomForestClassifier is. If using an Intel CPU, the optimizations provided by the scikit-learn-intelex patch might make setting ``extratrees=False`` worthwhile. If you are not using an Intel CPU, the scikit-learn-intelex library is not supported and ExtraTreesClassifier will be faster with similar performance. NOTE: If using scikit-learn-intelex, ``criterion`` must be set to "gini" and ``oob_score`` to False, as those parameters are not currently supported herein. Defaults to True.

        n_estimators (int, optional): The number of trees in the forest. Increasing this value can improve the fit, but at the cost of compute time and resources. Defaults to 100.

        criterion (str, optional): The function to measure the quality of a split. Supported values are "gini" for the Gini impurity and "entropy" for the information gain. Defaults to "gini".

        max_depth (int, optional): The maximum depth of the tree. If None, then nodes are expanded until all leaves are pure or until all leaves contain less than min_samples_split samples. Defaults to None.

        min_samples_split (int or float, optional): The minimum number of samples required to split an internal node. If value is an integer, then considers min_samples_split as the minimum number. If value is a floating point, then min_samples_split is a fraction and (min_samples_split * n_samples), rounded up to the nearest integer, are the minimum number of samples for each split. Defaults to 2.

        min_samples_leaf (int or float, optional): The minimum number of samples required to be at a leaf node. A split point at any depth will only be considered if it leaves at least ``min_samples_leaf`` training samples in each of the left and right branches. This may have the effect of smoothing the model, especially in regression. If value is an integer, then ``min_samples_leaf`` is the minimum number. If value is floating point, then ``min_samples_leaf`` is a fraction and ``int(min_samples_leaf * n_samples)`` is the minimum number of samples for each node. Defaults to 1.

        min_weight_fraction_leaf (float, optional): The minimum weighted fraction of the sum total of weights (of all the input samples) required to be at a leaf node. Samples have equal weight when sample_weight is not provided. Defaults to 0.0.

        max_features (str, int, float, or None, optional): The number of features to consider when looking for the best split. If int, then consider "max_features" features at each split. If float, then "max_features" is a fraction and ``int(max_features * n_samples)`` features are considered at each split. If "sqrt", then ``max_features=sqrt(n_features)``\. If "log2", then ``max_features=log2(n_features)``\. If None, then ``max_features=n_features``\. Defaults to "sqrt".

        max_leaf_nodes (int or None, optional): Grow trees with ``max_leaf_nodes`` in best-first fashion. Best nodes are defined as relative reduction in impurity. If None then unlimited number of leaf nodes. Defaults to None.

        min_impurity_decrease (float, optional): A node will be split if this split induces a decrease of the impurity greater than or equal to this value. See ``sklearn.ensemble.ExtraTreesClassifier`` documentation for more information. Defaults to 0.0.

        bootstrap (bool, optional): Whether bootstrap samples are used when building trees. If False, the whole dataset is used to build each tree. Defaults to False.

        oob_score (bool, optional): Whether to use out-of-bag samples to estimate the generalization score. Only available if ``bootstrap=True``\. Defaults to False.

        max_samples (int or float, optional): If bootstrap is True, the number of samples to draw from X to train each base estimator. If None (default), then draws ``X.shape[0] samples``\. if int, then draws ``max_samples`` samples. If float, then draws ``int(max_samples * X.shape[0] samples)`` with ``max_samples`` in the interval (0, 1). Defaults to None.


    Example:
        >>> data = GenotypeData(
        >>>     filename="test.str",
        >>>     filetype="auto",
        >>>     guidetree="test.tre",
        >>>     qmatrix_iqtree="test.iqtree"
        >>> )
        >>>
        >>> # Genetic Algorithm grid_params
        >>> grid_params = {
        >>>     "min_samples_leaf": Integer(1, 10),
        >>>     "max_depth": Integer(2, 110),
        >>> }
        >>>
        >>> rf = ImputeRandomForest(
        >>>     genotype_data=data,
        >>>     gridparams=grid_params,
        >>>     cv=5,
        >>>     gridsearch_method="genetic_algorithm",
        >>>     n_nearest_features=10,
        >>>     n_estimators=100,
        >>>     initial_strategy="phylogeny",
        >>> )
        >>>
        >>> rf_gtdata = rf.imputed
    """

    def __init__(
        self,
        genotype_data: Any,
        *,
        extratrees: bool = True,
        n_estimators: int = 100,
        criterion: str = "gini",
        max_depth: Optional[int] = None,
        min_samples_split: Union[int, float] = 2,
        min_samples_leaf: Union[int, float] = 1,
        min_weight_fraction_leaf: float = 0.0,
        max_features: Optional[Union[str, int, float]] = "sqrt",
        max_leaf_nodes: Optional[int] = None,
        min_impurity_decrease: float = 0.0,
        bootstrap: bool = False,
        oob_score: bool = False,
        max_samples: Optional[Union[int, float]] = None,
        **kwargs,
    ) -> None:
        all_kwargs = locals()
        kwargs.update(all_kwargs)
        kwargs.pop("self")
        kwargs.pop("genotype_data")

        self.extratrees = kwargs.pop("extratrees")

        if self.extratrees:
            self.clf = ExtraTreesClassifier

        elif os.environ["intelex"] == "True" and not self.extratrees:
            self.clf = RandomForestClassifier

            if kwargs["criterion"] != "gini":
                raise ValueError(
                    "criterion must be set to 'gini' if using the RandomForestClassifier with scikit-learn-intelex"
                )
            if kwargs["oob_score"]:
                raise ValueError(
                    "oob_score must be set to False if using the "
                    "RandomForestClassifier with scikit-learn-intelex"
                )
        else:
            self.clf = RandomForestClassifier

        self.clf_type = "classifier"

        super().__init__(genotype_data, self.clf, self.clf_type, **kwargs)


class ImputeXGBoost(SupervisedImputer):
    """Does XGBoost (Extreme Gradient Boosting) Iterative imputation of missing data. Iterative imputation uses the n_nearest_features to inform the imputation at each feature (i.e., SNP site), using the N most correlated features per site. The N most correlated features are drawn with probability proportional to correlation for each imputed target feature to ensure coverage of features throughout the imputation process.

    Args:
        genotype_data (GenotypeData object): GenotypeData instance that was used to read in the sequence data.

        n_estimators (int, optional): The number of boosting rounds. Increasing this value can improve the fit, but at the cost of compute time and RAM usage. Defaults to 100.

        max_depth (int, optional): Maximum tree depth for base learners. Defaults to 3.

        learning_rate (float, optional): Boosting learning rate (eta). Basically, it serves as a weighting factor for correcting new trees when they are added to the model. Typical values are between 0.1 and 0.3. Lower learning rates generally find the best optimum at the cost of requiring far more compute time and resources. Defaults to 0.1.

        booster (str, optional): Specify which booster to use. Possible values include "gbtree", "gblinear", and "dart". Defaults to "gbtree".

        gamma (float, optional): Minimum loss reduction required to make a further partition on a leaf node of the tree. Defaults to 0.0.

        min_child_weight (float, optional): Minimum sum of instance weight(hessian) needed in a child. Defaults to 1.0.

        max_delta_step (float, optional): Maximum delta step we allow each tree's weight estimation to be. Defaults to 0.0.

        subsample (float, optional): Subsample ratio of the training instance. Defaults to 1.0.

        colsample_bytree (float, optional): Subsample ratio of columns when constructing each tree. Defaults to 1.0.

        reg_lambda (float, optional): L2 regularization term on weights (xgb's lambda parameter). Defaults to 1.0.

        reg_alpha (float, optional): L1 regularization term on weights (xgb's alpha parameter). Defaults to 1.0.

    Example:
        >>> data = GenotypeData(
        >>>     filename="test.str",
        >>>     filetype="auto",
        >>>     guidetree="test.tre",
        >>>     qmatrix_iqtree="test.iqtree"
        >>> )
        >>>
        >>> # Genetic Algorithm grid_params
        >>> grid_params = {
        >>>     "learning_rate": Continuous(lower=0.01, upper=0.1),
        >>>     "max_depth": Integer(2, 110),
        >>> }
        >>>
        >>> xgb = ImputeXGBoost(
        >>>     genotype_data=data,
        >>>     gridparams=grid_params,
        >>>     cv=5,
        >>>     gridsearch_method="genetic_algorithm",
        >>>     n_nearest_features=10,
        >>>     n_estimators=100,
        >>>     initial_strategy="phylogeny",
        >>> )
        >>>
        >>> xgb_gtdata = xgb.imputed
    """

    def __init__(
        self,
        genotype_data: Any,
        *,
        n_estimators: int = 100,
        max_depth: int = 3,
        learning_rate: float = 0.1,
        booster: str = "gbtree",
        gamma: float = 0.0,
        min_child_weight: float = 1.0,
        max_delta_step: float = 0.0,
        subsample: float = 1.0,
        colsample_bytree: float = 1.0,
        reg_lambda: float = 1.0,
        reg_alpha: float = 0.0,
        **kwargs,
    ) -> None:
        # Get local variables into dictionary object
        all_kwargs = locals()
        kwargs.update(all_kwargs)

        self.clf_type = "classifier"
        self.clf = xgb.XGBClassifier
        kwargs["verbosity"] = kwargs["verbose"]

        super().__init__(genotype_data, self.clf, self.clf_type, **kwargs)


class ImputeVAE(UnsupervisedImputer):
    """Class to impute missing data using a Variational Autoencoder neural network model. For training, missing values are simulated and the model is trained on the simulated missing values. The real missing values are then predicted by the trained model. The strategy for simulating missing values can be set with the ``sim_strategy`` argument.

    Args:
        genotype_data (GenotypeData object): Input data initialized as GenotypeData object. Required positional argument.

        kl_beta (float, optional): Weight to apply to Kullback-Liebler divergence loss. If the latent distribution is not learned well, this weight can be adjusted to adjust how much KL divergence affects the total loss. Should be in the range [0, 1]. If set to 1.0, the KL loss is unweighted. If set to 0.0, the KL loss is negated entirely and does not affect the total loss. Defaults to 1.0.

    Example:
        >>> data = GenotypeData(
        >>>    filename="test.str",
        >>>    filetype="auto",
        >>>    guidetree="test.tre",
        >>>    qmatrix_iqtree="test.iqtree"
        >>> )
        >>>
        >>> vae = ImputeVAE(
        >>>     genotype_data=data,
        >>>     learning_rate=0.001,
        >>>     epochs=200,
        >>> )
        >>>
        >>> vae_gtdata = vae.imputed

    """

    def __init__(
        self,
        genotype_data,
        kl_beta=1.0,
        **kwargs,
    ):
        kwargs["kl_beta"] = kl_beta
        super().__init__(genotype_data, VAE, "classifier", **kwargs)


class ImputeStandardAutoEncoder(UnsupervisedImputer):
    """Class to impute missing data using a standard Autoencoder (SAE) neural network model. For training, missing values are simulated and the model is trained on the simulated missing values. The real missing values are then predicted by the trained model. The strategy for simulating missing values can be set with the ``sim_strategy`` argument.

    Args:
        genotype_data (GenotypeData object): Input data initialized as GenotypeData object. Required positional argument.

    Example:
        >>> data = GenotypeData(
        >>>    filename="test.str",
        >>>    filetype="auto",
        >>>    guidetree="test.tre",
        >>>    qmatrix_iqtree="test.iqtree"
        >>> )
        >>>
        >>> sae = ImputeStandardAutoEncoder(
        >>>     genotype_data=data,
        >>>     learning_rate=0.001,
        >>>     n_components=5,
        >>>     epochs=200,
        >>> )
        >>>
        >>> # Get the imputed data.
        >>> sae_gtdata = sae.imputed
    """

    def __init__(
        self,
        genotype_data,
        **kwargs,
    ):
        # Get local variables into dictionary object
        self.clf = SAE
        self.clf_type = "classifier"

        super().__init__(genotype_data, self.clf, self.clf_type, **kwargs)


class ImputeUBP(UnsupervisedImputer):
    """Class to impute missing data using an unsupervised backpropagation (UBP) neural network model. For training, missing values are simulated and the model is trained on the simulated missing values. The real missing values are then predicted by the trained model. The strategy for simulating missing values can be set with the ``sim_strategy`` argument.

    UBP [1]_ is an extension of NLPCA with the input being randomly generated and of reduced dimensionality that gets trained to predict the supplied output based on only known values. It then uses the trained model to predict missing values. However, in contrast to NLPCA, UBP trains the model over three phases. The first is a single layer perceptron used to refine the randomly generated input. The second phase is a multi-layer perceptron that uses the refined reduced-dimension data from the first phase as input. In the second phase, the model weights are refined but not the input. In the third phase, the model weights and the inputs are then refined.

    Args:
        genotype_data (GenotypeData object): Input data initialized as GenotypeData object. Required positional argument.

    Example:
        >>> data = GenotypeData(
        >>>    filename="test.str",
        >>>    filetype="auto",
        >>>    guidetree="test.tre",
        >>>    qmatrix_iqtree="test.iqtree"
        >>> )
        >>>
        >>> ubp = ImputeUBP(
        >>>     genotype_data=data,
        >>>     learning_rate=0.001,
        >>>     n_components=5
        >>> )
        >>>
        >>> # Get the imputed data.
        >>> ubp_gtdata = ubp.imputed

    References:
        .. [1] Gashler, M. S., Smith, M. R., Morris, R., & Martinez, T. (2016). Missing value imputation with unsupervised backpropagation. Computational Intelligence, 32(2), 196-215.
    """

    nlpca = False

    def __init__(
        self,
        genotype_data,
        **kwargs,
    ):
        self.clf = UBP
        self.clf_type = "classifier"
        if self.nlpca:
            self.clf.__name__ = "NLPCA"

        super().__init__(genotype_data, self.clf, self.clf_type, **kwargs)


class ImputeNLPCA(ImputeUBP):
    """Class to impute missing data using inverse non-linear principal component analysis (NLPCA) neural network models. For training, missing values are simulated and the model is trained on the simulated missing values. The real missing values are then predicted by the trained model. The strategy for simulating missing values can be set with the ``sim_strategy`` argument.

    NLPCA [2]_ trains randomly generated, reduced-dimensionality input to predict the correct output. In the case of imputation, the model is trained only on known values, and the trained model is then used to predict the missing values.

    Args:
        genotype_data (GenotypeData object): Input data initialized as GenotypeData object. Required positional argument.

    Example:
        >>> data = GenotypeData(
        >>>    filename="test.str",
        >>>    filetype="auto",
        >>>    guidetree="test.tre",
        >>>    qmatrix_iqtree="test.iqtree"
        >>> )
        >>>
        >>> nlpca = ImputeNLPCA(
        >>>     genotype_data=data,
        >>>     learning_rate=0.001,
        >>>     epochs=200
        >>> )
        >>>
        >>> nlpca_gtdata = nlpca.imputed

    References:

        .. [2] Scholz, M., Kaplan, F., Guy, C. L., Kopka, J., & Selbig, J. (2005). Non-linear PCA: a missing data approach. Bioinformatics, 21(20), 3887-3895.
    """

    nlpca = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import re
import sys
from itertools import product
from collections import Counter


def blacklist_missing(loci, threshold, iupac=False):
    blacklist = list()
    for i in range(0, len(loci)):
        alleles = expandLoci(loci[i], iupac=False)
        c = Counter(alleles)
        if float(c[-9] / sum(c.values())) > threshold:
            blacklist.append(i)
    return blacklist


def blacklist_maf(loci, threshold, iupac=False):
    blacklist = list()
    for i in range(0, len(loci)):
        alleles = expandLoci(loci[i], iupac=False)
        c = Counter(alleles)
        if len(c.keys()) <= 1:
            blacklist.append(i)
            continue
        else:
            minor_count = c.most_common(2)[1][1]
            if float(minor_count / sum(c.values())) < threshold:
                blacklist.append(i)
    return blacklist


def expandLoci(loc, iupac=False):
    """List of genotypes in 0-1-2 format."""
    ret = list()
    for i in loc:
        if not iupac:
            ret.extend(expand012(i))
        else:
            ret.extent(get_iupac_caseless(i))
    return ret


def expand012(geno):
    g = str(geno)
    if g == "0":
        return [0, 0]
    elif g == "1":
        return [0, 1]
    elif g == "2":
        return [1, 1]
    else:
        return [-9, -9]


def remove_items(all_list, bad_list):
    """Remove items from list using another list."""
    # using list comprehension to perform the task
    return [i for i in all_list if i not in bad_list]


def count_alleles(l, vcf=False):
    """Count how many total alleles there are.

    Args:
        l (List[str]): List of IUPAC or VCF-style (e.g. 0/1) genotypes.
        vcf (bool, optional): Whether genotypes are VCF or STRUCTURE-style. Defaults to False.

    Returns:
        int: Total number of alleles in l.
    """
    all_items = list()
    for i in l:
        if vcf:
            all_items.extend(i.split("/"))
        else:
            all_items.extend(get_iupac_caseless(i))
    all_items = remove_items(all_items, ["-9", "-", "N", -9])
    return len(set(all_items))


def get_major_allele(l, num=None, vcf=False):
    """Get most common alleles in list.

    Args:
        l (List[str]): List of genotypes for one sample.

        num (int, optional): Number of elements to return. Defaults to None.

        vcf (bool, optional): Alleles in VCF or STRUCTURE-style format. Defaults to False.

    Returns:
        list: Most common alleles in descending order.
    """
    all_items = list()
    for i in l:
        if vcf:
            all_items.extend(i.split("/"))
        else:
            all_items.extend(get_iupac_caseless(i))

    c = Counter(all_items)  # requires collections import

    # List of tuples with [(allele, count), ...] in order of
    # most to least common
    rets = c.most_common(num)

    # Returns two most common non-ambiguous bases
    # Makes sure the least common base isn't N or -9
    if vcf:
        return [x[0] for x in rets if x[0] != "-9"]
    else:
        return [x[0] for x in rets if x[0] in ["A", "T", "G", "C"]]


def get_iupac_caseless(char):
    """Split IUPAC code to two primary characters, assuming diploidy.

    Gives all non-valid ambiguities as N.

    Args:
        char (str): Base to expand into diploid list.

    Returns:
        List[str]: List of the two expanded alleles.
    """
    lower = False
    if char.islower():
        lower = True
        char = char.upper()
    iupac = {
        "A": ["A", "A"],
        "G": ["G", "G"],
        "C": ["C", "C"],
        "T": ["T", "T"],
        "N": ["N", "N"],
        "-": ["N", "N"],
        "R": ["A", "G"],
        "Y": ["C", "T"],
        "S": ["G", "C"],
        "W": ["A", "T"],
        "K": ["G", "T"],
        "M": ["A", "C"],
        "B": ["N", "N"],
        "D": ["N", "N"],
        "H": ["N", "N"],
        "V": ["N", "N"],
    }
    ret = iupac[char]
    if lower:
        ret = [c.lower() for c in ret]
    return ret


def get_iupac_full(char):
    """Split IUPAC code to all possible primary characters.

    Gives all ambiguities as "N".

    Args:
        char (str): Base to expaned into list.

    Returns:
        List[str]: List of the expanded alleles.
    """
    char = char.upper()
    iupac = {
        "A": ["A"],
        "G": ["G"],
        "C": ["C"],
        "T": ["T"],
        "N": ["A", "C", "T", "G"],
        "-": ["A", "C", "T", "G"],
        "R": ["A", "G"],
        "Y": ["C", "T"],
        "S": ["G", "C"],
        "W": ["A", "T"],
        "K": ["G", "T"],
        "M": ["A", "C"],
        "B": ["C", "G", "T"],
        "D": ["A", "G", "T"],
        "H": ["A", "C", "T"],
        "V": ["A", "C", "G"],
    }
    ret = iupac[char]
    return ret


def expandAmbiquousDNA(sequence):
    """Generator function to expand ambiguous sequences"""
    for i in product(*[get_iupac_caseless(j) for j in sequence]):
        yield ("".join(i))


def get_revComp_caseless(char):
    """Function to return reverse complement of a nucleotide, while preserving case."""
    lower = False
    if char.islower():
        lower = True
        char = char.upper()
    d = {
        "A": "T",
        "G": "C",
        "C": "G",
        "T": "A",
        "N": "N",
        "-": "-",
        "R": "Y",
        "Y": "R",
        "S": "S",
        "W": "W",
        "K": "M",
        "M": "K",
        "B": "V",
        "D": "H",
        "H": "D",
        "V": "B",
    }
    ret = d[char]
    if lower:
        ret = ret.lower()
    return ret


def reverseComplement(seq):
    """Function to reverse complement a sequence, with case preserved."""
    comp = []
    for i in (get_revComp_caseless(j) for j in seq):
        comp.append(i)
    return "".join(comp[::-1])


def simplifySeq(seq):
    """Function to simplify a sequence."""
    temp = re.sub("[ACGT]", "", (seq).upper())
    return temp.translate(str.maketrans("RYSWKMBDHV", "**********"))


def seqCounter(seq):
    """Returns dict of character counts"""
    d = {}
    d = {
        "A": 0,
        "N": 0,
        "-": 0,
        "C": 0,
        "G": 0,
        "T": 0,
        "R": 0,
        "Y": 0,
        "S": 0,
        "W": 0,
        "K": 0,
        "M": 0,
        "B": 0,
        "D": 0,
        "H": 0,
        "V": 0,
    }
    for c in seq:
        if c in d:
            d[c] += 1
    d["VAR"] = (
        d["R"]
        + d["Y"]
        + d["S"]
        + d["W"]
        + d["K"]
        + d["M"]
        + d["B"]
        + d["D"]
        + d["H"]
        + d["V"]
    )
    return d


def getFlankCounts(ref, x, y, dist):
    """Get vars, gaps, and N counts for flanking regions of a substring."""
    x2 = x - dist
    if x2 < 0:
        x2 = 0
    y2 = y + dist
    if y2 > len(ref):
        y2 = len(ref)
    flanks = ref[x2:x] + ref[y:y2]  # flanks = right + left flank
    counts = seqCounterSimple(simplifySeq(flanks))
    return counts


def seqCounterSimple(seq):
    """Get dict of character counts from a simplified consensus sequence."""
    d = {}
    d = {"N": 0, "-": 0, "*": 0}
    for c in seq:
        if c in d:
            d[c] += 1
    return d


def gc_counts(string):
    """Get GC content of a provided sequence."""
    new = re.sub("[GCgc]", "#", string)
    return sum(1 for c in new if c == "#")


def mask_counts(string):
    """Get counts of masked bases."""
    return sum(1 for c in string if c.islower())


def gc_content(string):
    """Get GC content as proportion."""
    new = re.sub("[GCgc]", "#", string)
    count = sum(1 for c in new if c == "#")
    return count / (len(string))


def mask_content(string):
    """Count number of lower case in a string."""
    count = sum(1 for c in string if c.islower())
    return count / (len(string))


def seqSlidingWindowString(seq, shift, width):
    """Generator to create sliding windows by slicing out substrings."""
    seqlen = len(seq)
    for i in range(0, seqlen, shift):
        if i + width > seqlen:
            j = seqlen
        else:
            j = i + width
        yield seq[i:j]
        if j == seqlen:
            break


def seqSlidingWindow(seq, shift, width):
    """Generator to create sliding windows by slicing out substrings."""
    seqlen = len(seq)
    for i in range(0, seqlen, shift):
        if i + width > seqlen:
            j = seqlen
        else:
            j = i + width
        yield [seq[i:j], i, j]
        if j == seqlen:
            break


def stringSubstitute(s, pos, c):
    """Fast way to replace single char in string.

    This way is a lot faster than doing it by making a list and subst in list.
    """
    return s[:pos] + c + s[pos + 1 :]


def listToSortUniqueString(l):
    """Get sorted unique string from list of chars.

    Args:
        l (List[str]): List of characters.

    Returns:
        List[str]: Sorted unique strings from list.
    """
    sl = sorted(set(l))
    return str("".join(sl))


def n_lower_chars(string):
    """Count number of lower case in a string."""
    return sum(1 for c in string if c.islower())


def countSlidingWindow(seq, shift, width):
    """Simplify a sequence to SNP, gaps, and Ns; get counts of sliding windows."""
    seq_temp = re.sub("[ACGT]", "", seq.upper())
    seq_norm = seq_temp.translate(str.maketrans("RYSWKMBDHV", "**********"))
    for i in windowSub(seq_norm, shift, width):
        # print(i)
        window_seq = "".join(i)
        seqCounterSimple(window_seq)


class slidingWindowGenerator:
    """Object for creating an iterable sliding window sampling."""

    # Need to come back and comment better...
    def __init__(self, seq, shift, width):
        self.__seq = seq
        self.__seqlen = len(self.__seq)
        self.__shift = shift
        self.__width = width
        self.__i = 0

    def __call__(self):
        self.__seqlen
        while self.__i < self.__seqlen:
            # print("i is ", self.__i, " : Base is ", self.__seq[self.__i]) #debug print
            if self.__i + self.__width > self.__seqlen:
                j = self.__seqlen
            else:
                j = self.__i + self.__width
            yield [self.__seq[self.__i : j], self.__i, j]
            if j == self.__seqlen:
                break



from setuptools import setup, find_packages
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


NAME = "pg-sui"
VERSION = "1.0.1"
AUTHORS = "Bradley T. Martin and Tyler K. Chafin"
AUTHOR_EMAIL = "evobio721@gmail.com"
MAINTAINER = "Bradley T. Martin"
DESCRIPTION = "Python machine and deep learning package to impute missing SNPs"
LONG_DESCRIPTION = open("README.md").read()

setup(
    name=NAME,
    version=VERSION,
    author=AUTHORS,
    author_email=AUTHOR_EMAIL,
    maintainer=MAINTAINER,
    maintainer_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/btmartin721/PG-SUI",
    project_urls={
        "Homepage": "https://github.com/btmartin721/PG-SUI",
        "Documentation": "https://pg-sui.readthedocs.io/en/latest/",
        "Source": "https://github.com/btmartin721/PG-SUI.git",
        "Bug Tracker": "https://github.com/btmartin721/PG-SUI/issues",
    },
    keywords=[
        "impute",
        "imputation",
        "imputer",
        "machine learning",
        "neural network",
        "deep learning",
        "python",
        "api",
        "vae",
        "autoencoder",
        "ubp",
        "nlpca",
        "population genomics",
        "unsupervised",
        "supervised",
        "IterativeImputer",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    license="GNU General Public License v3 (GPLv3)",
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "matplotlib",
        "seaborn",
        "jupyterlab",
        "scikit-learn>=1.0",
        "tqdm",
        "pandas",
        "numpy==1.24.3",
        "scipy",
        "xgboost",
        "tensorflow",
        "keras",
        "toytree",
        "sklearn-genetic-opt[all]>=0.6.0",
        "importlib-resources>=1.1.0",
        "pyvolve",
        "scikeras",
        "snpio",
        "urllib3>=1.26.7,<2.0.0",
        "typing-extensions<4.6.0",
    ],
    extras_require={
        "intel": ["scikit-learn-intelex"],
        "docs": ["sphinx<7", "sphinx-rtd-theme", "sphinx_autodoc_typehints"],
    },
    package_data={
        "pgsui": [
            "example_data/structure_files/*.str",
            "example_data/phylip_files/*.phy",
            "example_data/vcf_files/*",
            "example_data/popmaps/test.popmap",
            "example_data/trees/*",
        ]
    },
    include_package_data=True,
)



import math
import sys

import numpy as np
import tensorflow as tf


class CyclicalAnnealingCallback(tf.keras.callbacks.Callback):
    """Perform cyclical annealing with KL Divergence weights.

    The dynamically changing weight (beta) is multiplied with the KL Divergence loss.

    This process is supposed to improve the latent distribution sampling for the variational autoencoder model and eliminate the KL vanishing issue.

    Three types of cycle curves can be used that determine how the weight increases: 'linear', 'sigmoid', and 'cosine'..

    Code is adapted from: https://github.com/haofuml/cyclical_annealing

    The cyclical annealing process was first described in the following paper: https://aclanthology.org/N19-1021.pdf

    Args:
        n_iter (int): Number of iterations (epochs) being used in training.
        start (float, optional): Where to start cycles. Defaults to 0.0.
        stop (float, optional): Where to stop cycles. Defaults to 1.0.
        n_cycle (int, optional): How many cycles to use across all the epochs. Defaults to 4.
        ratio (float, optional): Ratio to determine proportion used to increase beta. Defaults to 0.5.
        schedule_type (str, optional): Type of curve to use for scheduler. Possible options include: 'linear', 'sigmoid', or 'cosine'. Defaults to 'linear'.

    """

    def __init__(
        self,
        n_iter,
        start=0.0,
        stop=1.0,
        n_cycle=4,
        ratio=0.5,
        schedule_type="linear",
    ):
        self.n_iter = n_iter
        self.start = start
        self.stop = stop
        self.n_cycle = n_cycle
        self.ratio = ratio
        self.schedule_type = schedule_type

        self.arr = None

    def on_train_begin(self, logs=None):
        """Executes on training begin.

        Here, the cycle curve is generated and stored as a class variable.
        """
        if self.schedule_type == "linear":
            cycle_func = self._linear_cycle_range
        elif self.schedule_type == "sigmoid":
            cycle_func = self._sigmoid_cycle_range
        elif self.schedule_type == "cosine":
            cycle_func = self._cosine_cycle_range
        else:
            raise ValueError(
                f"Invalid schedule_type value provided: {self.schedule_type}"
            )

        self.arr = cycle_func()

    def on_epoch_begin(self, epoch, logs=None):
        """Executes each time an epoch begins.

        Here, the new kl_beta weight is set.

        Args:
            epoch (int): Current epoch iteration.
            logs (None, optional): For compatibility. Not used. Defaults to None.
        """
        idx = epoch - 1
        new_weight = self.arr[idx]

        tf.keras.backend.set_value(self.model.kl_beta, new_weight)

    def _linear_cycle_range(self):
        """Get an array with a linear cycle curve ranging from 0 to 1 for n_iter epochs.

        The amount of time cycling and spent at 1.0 is determined by the ratio variable.

        Returns:
            numpy.ndarray: Linear cycle range.
        """
        L = np.ones(self.n_iter) * self.stop
        period = self.n_iter / self.n_cycle

        # Linear schedule
        step = (self.stop - self.start) / (
            period * self.ratio
        )  # linear schedule

        for c in range(self.n_cycle):
            v, i = self.start, 0
            while v <= self.stop and (int(i + c * period) < self.n_iter):
                L[int(i + c * period)] = v
                v += step
                i += 1

        return L

    def _sigmoid_cycle_range(self):
        """Get sigmoidal curve cycle ranging from 0 to 1 for n_iter epochs.

        The amount of time cycling and spent at 1.0 is determined by the ratio variable.

        Returns:
            numpy.ndarray: Sigmoidal cycle range.
        """
        L = np.ones(self.n_iter)
        period = self.n_iter / self.n_cycle
        step = (self.stop - self.start) / (
            period * self.ratio
        )  # step is in [0,1]

        for c in range(self.n_cycle):
            v, i = self.start, 0

            while v <= self.stop:
                L[int(i + c * period)] = 1.0 / (
                    1.0 + np.exp(-(v * 12.0 - 6.0))
                )
                v += step
                i += 1
        return L

    def _cosine_cycle_range(self):
        """Get cosine curve cycle ranging from 0 to 1 for n_iter epochs.

        The amount of time cycling and spent at 1.0 is determined by the ratio variable.

        Returns:
            numpy.ndarray: Cosine cycle range.
        """
        L = np.ones(self.n_iter)
        period = self.n_iter / self.n_cycle
        step = (self.stop - self.start) / (
            period * self.ratio
        )  # step is in [0,1]

        for c in range(self.n_cycle):
            v, i = self.start, 0

            while v <= self.stop:
                L[int(i + c * period)] = 0.5 - 0.5 * math.cos(v * math.pi)
                v += step
                i += 1
        return L


class VAECallbacks(tf.keras.callbacks.Callback):
    """Custom callbacks to use with subclassed VAE Keras model.

    Requires y, missing_mask, and sample_weight to be input variables to be properties with setters in the subclassed model.
    """

    def __init__(self):
        self.indices = None

    def on_epoch_begin(self, epoch, logs=None):
        """Shuffle input and target at start of epoch."""
        y = self.model.y.copy()
        missing_mask = self.model.missing_mask
        sample_weight = self.model.sample_weight

        n_samples = len(y)
        self.indices = np.arange(n_samples)
        np.random.shuffle(self.indices)

        self.model.y = y[self.indices]
        self.model.missing_mask = missing_mask[self.indices]

        if sample_weight is not None:
            self.model.sample_weight = sample_weight[self.indices]

    def on_train_batch_begin(self, batch, logs=None):
        """Get batch index."""
        self.model.batch_idx = batch

    def on_epoch_end(self, epoch, logs=None):
        """Unsort the row indices."""
        unshuffled = np.argsort(self.indices)

        self.model.y = self.model.y[unshuffled]
        self.model.missing_mask = self.model.missing_mask[unshuffled]

        if self.model.sample_weight is not None:
            self.model.sample_weight = self.model.sample_weight[unshuffled]


class UBPCallbacks(tf.keras.callbacks.Callback):
    """Custom callbacks to use with subclassed NLPCA/ UBP Keras models.

    Requires y, missing_mask, V_latent, and sample_weight to be input variables to be properties with setters in the subclassed model.
    """

    def __init__(self):
        self.indices = None

    def on_epoch_begin(self, epoch, logs=None):
        """Shuffle input and target at start of epoch."""
        y = self.model.y.copy()
        missing_mask = self.model.missing_mask
        sample_weight = self.model.sample_weight

        n_samples = len(y)
        self.indices = np.arange(n_samples)
        np.random.shuffle(self.indices)

        self.model.y = y[self.indices]
        self.model.V_latent = self.model.V_latent[self.indices]
        self.model.missing_mask = missing_mask[self.indices]

        if sample_weight is not None:
            self.model.sample_weight = sample_weight[self.indices]

    def on_train_batch_begin(self, batch, logs=None):
        """Get batch index."""
        self.model.batch_idx = batch

    def on_epoch_end(self, epoch, logs=None):
        """Unsort the row indices."""
        unshuffled = np.argsort(self.indices)

        self.model.y = self.model.y[unshuffled]
        self.model.V_latent = self.model.V_latent[unshuffled]
        self.model.missing_mask = self.model.missing_mask[unshuffled]

        if self.model.sample_weight is not None:
            self.model.sample_weight = self.model.sample_weight[unshuffled]


class UBPEarlyStopping(tf.keras.callbacks.Callback):
    """Stop training when the loss is at its min, i.e. the loss stops decreasing.

    Args:
        patience (int, optional): Number of epochs to wait after min has been hit. After this
        number of no improvement, training stops. Defaults to 0.

        phase (int, optional): Current UBP Phase. Defaults to 3.
    """

    def __init__(self, patience=0, phase=3):
        super(UBPEarlyStopping, self).__init__()
        self.patience = patience
        self.phase = phase

        # best_weights to store the weights at which the minimum loss occurs.
        self.best_weights = None

        # In UBP, the input gets refined during training.
        # So we have to revert it too.
        self.best_input = None

    def on_train_begin(self, logs=None):
        # The number of epoch it has waited when loss is no longer minimum.
        self.wait = 0
        # The epoch the training stops at.
        self.stopped_epoch = 0
        # Initialize the best as infinity.
        self.best = np.Inf

    def on_epoch_end(self, epoch, logs=None):
        current = logs.get("loss")
        if np.less(current, self.best):
            self.best = current
            self.wait = 0
            # Record the best weights if current results is better (less).
            self.best_weights = self.model.get_weights()

            if self.phase != 2:
                # Only refine input in phase 2.
                self.best_input = self.model.V_latent
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.model.stop_training = True
                self.model.set_weights(self.best_weights)

                if self.phase != 2:
                    self.model.V_latent = self.best_input



import sys

import numpy as np

from sklearn.metrics import (
    roc_curve,
    auc,
    accuracy_score,
    hamming_loss,
    make_scorer,
    precision_recall_curve,
    average_precision_score,
    multilabel_confusion_matrix,
    f1_score,
    roc_auc_score,
)

from sklearn.preprocessing import label_binarize

try:
    from ..impute.unsupervised.neural_network_methods import (
        NeuralNetworkMethods,
    )
except (ModuleNotFoundError, ValueError, ImportError):
    from impute.unsupervised.neural_network_methods import NeuralNetworkMethods


class Scorers:
    @staticmethod
    def compute_roc_auc_micro_macro(
        y_true, y_pred, num_classes=3, binarize_pred=True
    ):
        """Compute ROC curve with AUC scores.

        ROC (Receiver Operating Characteristic) curves and AUC (area under curve) scores are computed per-class and for micro and macro averages.

        Args:
            y_true (numpy.ndarray): Ravelled numpy array of shape (n_samples * n_features,). y_true should be integer-encoded.

            y_pred (numpy.ndarray): Ravelled numpy array of shape (n_samples * n_features,). y_pred should be probabilities.

            num_classes (int, optional): How many classes to use. Defaults to 3.

            binarize_pred (bool, optional): Whether to binarize y_pred. If False, y_pred should be probabilities of each class. Defaults to True.

        Returns:
            Dict[str, Any]: Dictionary with true and false positive rates along probability threshold curve per class, micro and macro tpr and fpr curves averaged across classes, and AUC scores per-class and for micro and macro averages.
        """
        cats = range(num_classes)

        # Get only classes that appear in y_true.
        classes = [i for i in cats if i in y_true]

        # Binarize the output for use with ROC-AUC.
        y_true_bin = label_binarize(y_true, classes=cats)

        if binarize_pred:
            y_pred_bin = label_binarize(y_pred, classes=cats)
        else:
            y_pred_bin = y_pred

        for i in range(y_true_bin.shape[1]):
            if i not in classes:
                y_true_bin = np.delete(y_true_bin, i, axis=-1)
                y_pred_bin = np.delete(y_pred_bin, i, axis=-1)

        n_classes = len(classes)

        # Compute ROC curve and ROC area for each class.
        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        for i, c in enumerate(classes):
            fpr[c], tpr[c], _ = roc_curve(y_true_bin[:, i], y_pred_bin[:, i])
            roc_auc[c] = auc(fpr[c], tpr[c])

        # Compute micro-average ROC curve and ROC area.
        fpr["micro"], tpr["micro"], _ = roc_curve(
            y_true_bin.ravel(), y_pred_bin.ravel()
        )

        roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

        # Aggregate all false positive rates
        all_fpr = np.unique(np.concatenate([fpr[i] for i in classes]))

        # Then interpolate all ROC curves at these points.
        mean_tpr = np.zeros_like(all_fpr)
        for i in classes:
            mean_tpr += np.interp(all_fpr, fpr[i], tpr[i])

        # Finally, average it and compute AUC.
        mean_tpr /= n_classes

        fpr["macro"] = all_fpr
        tpr["macro"] = mean_tpr

        roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

        roc_auc["fpr_macro"] = fpr["macro"]
        roc_auc["tpr_macro"] = tpr["macro"]
        roc_auc["fpr_micro"] = fpr["micro"]
        roc_auc["tpr_micro"] = tpr["micro"]

        for i in classes:
            roc_auc[f"fpr_{i}"] = fpr[i]
            roc_auc[f"tpr_{i}"] = tpr[i]

        return roc_auc

    @staticmethod
    def compute_pr(y_true, y_pred, use_int_encodings=False, num_classes=4):
        """Compute precision-recall curve with Average Precision scores.

        PR and AP scores are computed per-class and for micro and macro averages.

        Args:
            y_true (numpy.ndarray): Ravelled numpy array of shape (n_samples * n_features,).

            y_pred (numpy.ndarray): Ravelled numpy array of shape (n_samples * n_features,). y_pred should be integer-encoded.

            use_int_encodings (bool, optional): Whether the imputer model is a neural network model. Defaults to False.

            num_classes (int, optional): How many classes to use. Defaults to 3.

         Returns:
            Dict[str, Any]: Dictionary with precision and recall curves per class and micro and macro averaged across classes, plus AP scores per-class and for micro and macro averages.
        """
        cats = range(num_classes)

        is_multiclass = True if num_classes != 4 else False

        # Get only classes that appear in y_true.
        classes = [i for i in cats if i in y_true]

        # Binarize the output for use with ROC-AUC.
        y_true_bin = label_binarize(y_true, classes=cats)
        y_pred_proba_bin = y_pred

        if is_multiclass:
            for i in range(y_true_bin.shape[1]):
                if i not in classes:
                    y_true_bin = np.delete(y_true_bin, i, axis=-1)
                    y_pred_proba_bin = np.delete(y_pred_proba_bin, i, axis=-1)

        nn = NeuralNetworkMethods()
        if len(y_true.shape) == 1 or y_true.shape[1] != num_classes:
            y_true = label_binarize(y_true, classes=cats)

        # Ensure y_pred_012 is in the multilabel format
        if use_int_encodings:
            y_pred_012 = nn.decode_masked(
                y_true,
                y_pred_proba_bin,
                return_multilab=True,  # Ensure multilabel format is returned
            )
        else:
            y_pred_012 = nn.decode_masked(
                y_true,
                y_pred_proba_bin,
                is_multiclass=False,
                return_int=False,
                return_multilab=True,
            )

        # Make confusion matrix to get true negatives and true positives.
        mcm = multilabel_confusion_matrix(y_true, y_pred_012)

        tn = np.sum(mcm[:, 0, 0])
        tn /= num_classes

        tp = np.sum(mcm[:, 1, 1])
        tp /= num_classes

        baseline = tp / (tn + tp)

        precision = dict()
        recall = dict()
        average_precision = dict()

        for i, c in enumerate(classes):
            precision[c], recall[c], _ = precision_recall_curve(
                y_true_bin[:, i], y_pred_proba_bin[:, i]
            )
            average_precision[c] = average_precision_score(
                y_true_bin[:, i], y_pred_proba_bin[:, i]
            )

        # A "micro-average": quantifying score on all classes jointly.
        precision["micro"], recall["micro"], _ = precision_recall_curve(
            y_true_bin.ravel(), y_pred_proba_bin.ravel()
        )

        average_precision["micro"] = average_precision_score(
            y_true_bin, y_pred_proba_bin, average="micro"
        )

        average_precision["macro"] = average_precision_score(
            y_true_bin, y_pred_proba_bin, average="macro"
        )

        average_precision["weighted"] = average_precision_score(
            y_true_bin, y_pred_proba_bin, average="weighted"
        )

        if use_int_encodings:
            y_pred_012 = (
                nn.decode_masked(
                    y_true_bin,
                    y_pred_proba_bin,
                    return_multilab=True,
                    predict_still_missing=False,
                ),
            )

        f1 = f1_score(y_true_bin, y_pred_012, average="macro")
        f1_weighted = f1_score(y_true_bin, y_pred_012, average="weighted")

        # Aggregate all recalls
        all_recall = np.unique(np.concatenate([recall[i] for i in classes]))

        # Then interpolate all PR curves at these points.
        mean_precision = np.zeros_like(all_recall)
        for i in classes:
            mean_precision += np.interp(all_recall, precision[i], recall[i])

        # Finally, average it and compute AUC.
        mean_precision /= num_classes

        recall["macro"] = all_recall
        precision["macro"] = mean_precision

        results = dict()

        results["micro"] = average_precision["micro"]
        results["macro"] = average_precision["macro"]
        results["f1_score"] = f1
        results["f1_score_weighted"] = f1_weighted
        results["recall_macro"] = all_recall
        results["precision_macro"] = mean_precision
        results["recall_micro"] = recall["micro"]
        results["precision_micro"] = precision["micro"]

        for i in classes:
            results[f"recall_{i}"] = recall[i]
            results[f"precision_{i}"] = precision[i]
            results[i] = average_precision[i]
        results["baseline"] = baseline

        return results

    @staticmethod
    def check_if_tuple(y_pred):
        """Checks if y_pred is a tuple and if so, returns the first element of the tuple."""
        if isinstance(y_pred, tuple):
            y_pred = y_pred[0]
        return y_pred

    @staticmethod
    def accuracy_scorer(y_true, y_pred, **kwargs):
        """Get accuracy score for grid search.

        If provided, only calculates score where missing_mask is True (i.e., data were missing). This is so that users can simulate missing data for known values, and then the predictions for only those known values can be evaluated.

        Args:
            y_true (numpy.ndarray): 012-encoded true target values.

            y_pred (tensorflow.EagerTensor): Predictions from model as probabilities. They must first be decoded to use with accuracy_score.

            kwargs (Any): Keyword arguments to use with scorer. Supported options include ``missing_mask`` and ``testing``\.

        Returns:
            float: Metric score by comparing y_true and y_pred.
        """
        missing_mask = kwargs.get("missing_mask")

        y_pred = Scorers.check_if_tuple(y_pred)

        y_true_masked = y_true[missing_mask]
        y_pred_masked = y_pred[missing_mask]

        nn = NeuralNetworkMethods()
        y_pred_masked_decoded = nn.decode_masked(
            y_true_masked, y_pred_masked, predict_still_missing=False
        )

        return accuracy_score(y_true_masked, y_pred_masked_decoded)

    @staticmethod
    def hamming_scorer(y_true, y_pred, **kwargs):
        """Get Hamming score for grid search.

        If provided, only calculates score where missing_mask is True (i.e., data were missing). This is so that users can simulate missing data for known values, and then the predictions for only those known values can be evaluated.

        Args:
            y_true (numpy.ndarray): 012-encoded true target values.

            y_pred (tensorflow.EagerTensor): Predictions from model as probabilities. They must first be decoded to use with hamming_scorer.

            kwargs (Any): Keyword arguments to use with scorer. Supported options include ``missing_mask`` and ``testing``\.

        Returns:
            float: Metric score by comparing y_true and y_pred.
        """
        missing_mask = kwargs.get("missing_mask")

        y_pred = Scorers.check_if_tuple(y_pred)

        y_true_masked = y_true[missing_mask]
        y_pred_masked = y_pred[missing_mask]

        nn = NeuralNetworkMethods()
        y_pred_masked_decoded = nn.decode_masked(
            y_true_masked,
            y_pred_masked,
            predict_still_missing=False,
        )

        return hamming_loss(y_true_masked, y_pred_masked_decoded)

    @staticmethod
    def compute_metric(
        y_true, y_pred, metric_type, scoring_function, **kwargs
    ):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)

        num_classes = kwargs.get("num_classes", 4)
        cats = range(num_classes)

        y_true_bin = label_binarize(y_true, classes=cats)

        if scoring_function == "roc_auc":
            return roc_auc_score(
                y_true_bin, y_pred, multi_class="ovr", average=metric_type
            )
        elif scoring_function == "f1":
            is_multiclass = num_classes != 4
            y_pred_proba_bin = y_pred
            nn = NeuralNetworkMethods()
            y_pred_bin = nn.decode_masked(
                y_true_bin,
                y_pred_proba_bin,
                is_multiclass=is_multiclass,
                return_int=False,
                return_multilab=True,
            )
            return f1_score(
                y_true_bin, y_pred_bin, average=metric_type, zero_division=0
            )
        elif scoring_function == "average_precision":
            return average_precision_score(
                y_true_bin, y_pred, average=metric_type
            )
        else:
            raise ValueError(
                f"Unsupported scoring function: {scoring_function}"
            )

    @staticmethod
    def compute_score(y_true, y_pred, metric_type, scoring_function, **kwargs):
        missing_mask = kwargs.get("missing_mask")
        y_pred = Scorers.check_if_tuple(y_pred)

        y_true_masked = y_true[missing_mask]
        y_pred_masked = y_pred[missing_mask]

        return Scorers.compute_metric(
            y_true_masked,
            y_pred_masked,
            metric_type=metric_type,
            scoring_function=scoring_function,
            **kwargs,
        )

    @classmethod
    def make_multimetric_scorer(
        cls, metrics, missing_mask, num_classes=4, testing=False
    ):
        if isinstance(metrics, str):
            metrics = [metrics]

        metric_map = {
            "roc_auc_macro": ("macro", "roc_auc"),
            "roc_auc_micro": ("micro", "roc_auc"),
            "roc_auc_weighted": ("weighted", "roc_auc"),
            "f1_micro": ("micro", "f1"),
            "f1_macro": ("macro", "f1"),
            "f1_weighted": ("weighted", "f1"),
            "average_precision_macro": ("macro", "average_precision"),
            "average_precision_micro": ("micro", "average_precision"),
            "average_precision_weighted": ("weighted", "average_precision"),
        }

        default_params = {
            "missing_mask": missing_mask,
            "num_classes": num_classes,
            "testing": testing,
        }

        scorers = dict()
        for item in metrics:
            item = item.lower()

            if item in metric_map:
                metric_type, scoring_function = metric_map[item]
                params = default_params.copy()
                scorers[item] = make_scorer(
                    Scorers.compute_score,
                    metric_type=metric_type,
                    scoring_function=scoring_function,
                    **params,
                )
            elif item == "accuracy":
                scorers[item] = make_scorer(
                    cls.accuracy_scorer, **default_params
                )
            elif item == "hamming":
                scorers[item] = make_scorer(
                    cls.hamming_scorer, **default_params
                )
            else:
                raise ValueError(f"Unsupported metric: {item}")

        return scorers

    @staticmethod
    def scorer(y_true, y_pred, **kwargs):
        # Get missing mask if provided.
        # Otherwise default is all missing values (array all True).
        missing_mask = kwargs.get("missing_mask")
        nn_model = kwargs.get("nn_model", True)
        num_classes = kwargs.get("num_classes", 3)
        testing = kwargs.get("testing", False)

        is_multiclass = True if num_classes != 4 else False

        if nn_model:
            nn = NeuralNetworkMethods()

        # VAE has tuple output.
        if isinstance(y_pred, tuple):
            y_pred = y_pred[0]

        y_true_masked = y_true[missing_mask]
        y_pred_masked = y_pred[missing_mask]

        roc_auc = Scorers.compute_roc_auc_micro_macro(
            y_true_masked,
            y_pred_masked,
            num_classes=num_classes,
            binarize_pred=False,
        )

        pr_ap = Scorers.compute_pr(
            y_true_masked,
            y_pred_masked,
            num_classes=num_classes,
        )

        acc = accuracy_score(
            y_true_masked,
            nn.decode_masked(
                y_true_masked,
                y_pred_masked,
                is_multiclass=is_multiclass,
                return_int=True,
            ),
        )
        ham = hamming_loss(
            y_true_masked,
            nn.decode_masked(
                y_true_masked,
                y_pred_masked,
                is_multiclass=is_multiclass,
                return_int=True,
            ),
        )

        if testing:
            y_pred_masked_decoded = nn.decode_masked(
                y_true_masked,
                y_pred_masked,
                is_multiclass=is_multiclass,
                return_int=True,
            )

            bin_mapping = [np.array2string(x) for x in y_pred_masked]

            with open("genotype_dist.csv", "w") as fout:
                fout.write(
                    "site,prob_vector,imputed_genotype,expected_genotype\n"
                )
                for i, (yt, yp, ypd) in enumerate(
                    zip(y_true_masked, bin_mapping, y_pred_masked_decoded)
                ):
                    fout.write(f"{i},{yp},{ypd},{yt}\n")
            # np.set_printoptions(threshold=np.inf)
            # print(y_true_masked)
            # print(y_pred_masked_decoded)

        metrics = dict()
        metrics["accuracy"] = acc
        metrics["roc_auc"] = roc_auc
        metrics["precision_recall"] = pr_ap
        metrics["hamming"] = ham

        return metrics
